import os
import json
import shutil

from crds.bestrefs import BestrefsScript
from crds.tests import test_config

BESTREFS = {
        "FLATFILE": 'x5v1944hl_flat.fits',
        "BADTTAB" : "NOT FOUND n/a",
        "BRFTAB" : "UNDEFINED",
        "TRACETAB" : "NOT FOUND some other error",
        }

BESTREFS_DEFINED = {
        "FLATFILE": 'x5v1944hl_flat.fits',
        }

BESTREFS_NA = {
        "FLATFILE" : "NOT FOUND n/a",
        }

BESTREFS_UNDEFINED = {
        "FLATFILE" : "UNDEFINED",
        }

BESTREFS_OMITTED= {
        }

BESTREFS_ERROR = {
    "FLATFILE" : "NOT FOUND some other error",
    }

"""Special case tests verify behavior in various configurations:

--regression
--na-differences-matter
--undefined-differences-matter

"""

def dt_bestrefs_na_undefined_single_ctx_defaults():
    """
    >>> old_state = test_config.setup()

    Instantiate a BestrefsScript with default settings and dummy parameters,  simulating pipeline bestrefs defaults:

    >>> script = BestrefsScript(argv="crds.bestrefs --load-pickles data/bestrefs.special.json --new-context hst_0315.pmap")
    >>> script.complex_init()
    CRDS - INFO - Loading file 'data/bestrefs.special.json'
    CRDS - INFO - Loaded 1 datasets from file 'data/bestrefs.special.json' completely replacing existing headers.
    CRDS - INFO - No comparison context or source comparison requested.
    True

    Skip ahead past bestrefs (faked) into special values results evaluation.

    First a postitive result:  OK,  lowercase-original-result,  uppercase-final-result-for-update

    >>> script.handle_na_and_not_found("New:", BESTREFS, "LA9K03C3Q", "COS", "FLATFILE")
    (True, 'X5V1944HL_FLAT.FITS')

    Second a formal N/A result:  OK,  bestrefs string,  final update value

    >>> script.handle_na_and_not_found("New:", BESTREFS, "LA9K03C3Q", "COS", "BADTTAB")
    (True, 'N/A')

    An explicit response of UNDEFINED,  by default converted to N/A for update,  considered OK

    >>> script.handle_na_and_not_found("New:", BESTREFS, "LA9K03C3Q", "COS", "BRFTAB")
    (True, 'N/A')

    An implicit response of UNDEFINED, i.e. OMIT, also coverted to N/A for update
    XTRACTAB intentionally omitted from BESTREFS

    >>> script.handle_na_and_not_found("New:", BESTREFS, "LA9K03C3Q", "COS", "XTRACTAB")
    (True, 'N/A')

    An explicit error response from CRDS,  not OK

    >>> script.handle_na_and_not_found("New:", BESTREFS, "LA9K03C3Q", "COS", "TRACETAB")
    CRDS - ERROR - instrument='COS' type='TRACETAB' data='LA9K03C3Q' ::  New: Bestref FAILED:  some other error
    (False, 'NOT FOUND SOME OTHER ERROR')

    >>> test_config.cleanup(old_state)

    """

def dt_bestrefs_na_undefined_single_ctx_undefined_matters():
    """
    >>> old_state = test_config.setup()

    Instantiate a BestrefsScript with default settings and dummy parameters,  simulating pipeline bestrefs defaults:

    >>> script = BestrefsScript(argv="crds.bestrefs --load-pickles data/bestrefs.special.json --undefined-differences-matter --new-context hst_0315.pmap")
    >>> script.complex_init()
    CRDS - INFO - Loading file 'data/bestrefs.special.json'
    CRDS - INFO - Loaded 1 datasets from file 'data/bestrefs.special.json' completely replacing existing headers.
    CRDS - INFO - No comparison context or source comparison requested.
    True

    Skip ahead past bestrefs (faked) into special values results evaluation.

    First a postitive result:  OK,  lowercase-original-result,  uppercase-final-result-for-update

    >>> script.handle_na_and_not_found("New:", BESTREFS, "LA9K03C3Q", "COS", "FLATFILE")
    (True, 'X5V1944HL_FLAT.FITS')

    Second a formal N/A result:  OK,  bestrefs string,  final update value

    >>> script.handle_na_and_not_found("New:", BESTREFS, "LA9K03C3Q", "COS", "BADTTAB")
    (True, 'N/A')

    An explicit response of UNDEFINED,  by default converted to N/A for update,  not OK since -undefined-differences-matter

    >>> script.handle_na_and_not_found("New:", BESTREFS, "LA9K03C3Q", "COS", "BRFTAB")
    CRDS - ERROR - instrument='COS' type='BRFTAB' data='LA9K03C3Q' ::  New: No match found: 'UNDEFINED'
    (False, 'UNDEFINED')

    An implicit response of UNDEFINED, i.e. OMIT, also coverted to N/A for update, not OK since -undefined-differences-matter

    >>> script.handle_na_and_not_found("New:", BESTREFS, "LA9K03C3Q", "COS", "XTRACTAB")
    CRDS - ERROR - instrument='COS' type='XTRACTAB' data='LA9K03C3Q' ::  New: No match found: 'UNDEFINED'
    (False, 'UNDEFINED')

    An explicit error response from CRDS,  not OK

    >>> script.handle_na_and_not_found("New:", BESTREFS, "LA9K03C3Q", "COS", "TRACETAB")
    CRDS - ERROR - instrument='COS' type='TRACETAB' data='LA9K03C3Q' ::  New: Bestref FAILED:  some other error
    (False, 'NOT FOUND SOME OTHER ERROR')

    >>> test_config.cleanup(old_state)

    """

def dt_bestrefs_na_undefined_single_ctx_na_matters():
    """
    >>> old_state = test_config.setup()

    Instantiate a BestrefsScript with default settings and dummy parameters,  simulating pipeline bestrefs defaults:

    >>> script = BestrefsScript(argv="crds.bestrefs --load-pickles data/bestrefs.special.json --na-differences-matter --new-context hst_0315.pmap")
    >>> script.complex_init()
    CRDS - INFO - Loading file 'data/bestrefs.special.json'
    CRDS - INFO - Loaded 1 datasets from file 'data/bestrefs.special.json' completely replacing existing headers.
    CRDS - INFO - No comparison context or source comparison requested.
    True

    Skip ahead past bestrefs (faked) into special values results evaluation.

    First a postitive result:  OK,  lowercase-original-result,  uppercase-final-result-for-update

    >>> script.handle_na_and_not_found("New:", BESTREFS, "LA9K03C3Q", "COS", "FLATFILE")
    (True, 'X5V1944HL_FLAT.FITS')

    Second a formal N/A result:  OK,  bestrefs string,  final update value

    >>> script.handle_na_and_not_found("New:", BESTREFS, "LA9K03C3Q", "COS", "BADTTAB")
    (True, 'N/A')

    An explicit response of UNDEFINED,  by default converted to N/A for update,  considered OK

    >>> script.handle_na_and_not_found("New:", BESTREFS, "LA9K03C3Q", "COS", "BRFTAB")
    (True, 'N/A')

    An implicit response of UNDEFINED, i.e. OMIT, also coverted to N/A for update

    >>> script.handle_na_and_not_found("New:", BESTREFS, "LA9K03C3Q", "COS", "XTRACTAB")
    (True, 'N/A')

    An explicit error response from CRDS, not OK

    >>> script.handle_na_and_not_found("New:", BESTREFS, "LA9K03C3Q", "COS", "TRACETAB")
    CRDS - ERROR - instrument='COS' type='TRACETAB' data='LA9K03C3Q' ::  New: Bestref FAILED:  some other error
    (False, 'NOT FOUND SOME OTHER ERROR')

    >>> test_config.cleanup(old_state)

    """

kinds = ["DEFINED", "NA", "UNDEFINED", "OMITTED", "ERROR"]

old_state = test_config.setup()

SCRIPT_UNDEFMATTERS_NAMATTERS  = BestrefsScript(argv="crds.bestrefs --load-pickles data/bestrefs.special.json --new-context hst_0315.pmap --undefined-differences-matter --na-differences-matter")
SCRIPT_UNDEFMATTERS_NAMATTERS.complex_init()

SCRIPT_UNDEFOK_NAMATTERS  = BestrefsScript(argv="crds.bestrefs --load-pickles data/bestrefs.special.json --new-context hst_0315.pmap --na-differences-matter")
SCRIPT_UNDEFOK_NAMATTERS.complex_init()

SCRIPT_UNDEFOK_NAOK  = BestrefsScript(argv="crds.bestrefs --load-pickles data/bestrefs.special.json --new-context hst_0315.pmap")
SCRIPT_UNDEFOK_NAOK.complex_init()

SCRIPT_UNDEFMATTERS_NAOK  = BestrefsScript(argv="crds.bestrefs --load-pickles data/bestrefs.special.json --new-context hst_0315.pmap --undefined-differences-matter")
SCRIPT_UNDEFMATTERS_NAOK.complex_init()

test_config.cleanup(old_state)
del old_state

def generate_comparisons(undef_matters, na_matters):
    for kind1 in kinds:
        for kind2 in kinds:
            bestrefs1 = globals()["BESTREFS_" + kind1]
            bestrefs2 = globals()["BESTREFS_" + kind2]
            undef = "undefmatters" if undef_matters else "undefok"
            na = "namatters" if na_matters else "naok"
            print("""

def dt_compare_bestrefs_%s_%s_%s_%s():
    '''
    >>> old_state = test_config.setup()

    >>> script_%s_%s.compare_bestrefs('%s', '%s', %s, %s)

    >>> test_config.cleanup(old_state)
    ''' """ % (

       kind1.lower(),
       kind2.lower(),
       undef,
       na,

       undef, na,

       "COS","LA9K03C3Q", bestrefs1, bestrefs2))

def generate_all():
    for undef_matters in [False, True]:
        for na_matters in [False, True]:
            generate_comparisons(undef_matters, na_matters)

# ==================================================================================

def main():
    """Run module tests,  for now just doctests only."""
    from crds.tests import test_special, tstmod
    return tstmod(test_special)

if __name__ == "__main__":
    print(main())

# ==================================================================================

def dt_compare_bestrefs_defined_defined_undefok_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'x5v1944hl_flat.fits'}, {'FLATFILE': 'x5v1944hl_flat.fits'})
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_defined_na_undefok_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'x5v1944hl_flat.fits'}, {'FLATFILE': 'NOT FOUND n/a'})
    [UpdateTuple(instrument='COS', filekind='flatfile', old_reference='X5V1944HL_FLAT.FITS', new_reference='N/A')]

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_defined_undefined_undefok_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'x5v1944hl_flat.fits'}, {'FLATFILE': 'UNDEFINED'})
    [UpdateTuple(instrument='COS', filekind='flatfile', old_reference='X5V1944HL_FLAT.FITS', new_reference='N/A')]

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_defined_omitted_undefok_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'x5v1944hl_flat.fits'}, {})
    [UpdateTuple(instrument='COS', filekind='flatfile', old_reference='X5V1944HL_FLAT.FITS', new_reference='N/A')]

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_defined_error_undefok_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'x5v1944hl_flat.fits'}, {'FLATFILE': 'NOT FOUND some other error'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: Bestref FAILED:  some other error
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_na_defined_undefok_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND n/a'}, {'FLATFILE': 'x5v1944hl_flat.fits'})
    [UpdateTuple(instrument='COS', filekind='flatfile', old_reference='N/A', new_reference='X5V1944HL_FLAT.FITS')]

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_na_na_undefok_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND n/a'}, {'FLATFILE': 'NOT FOUND n/a'})
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_na_undefined_undefok_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND n/a'}, {'FLATFILE': 'UNDEFINED'})
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_na_omitted_undefok_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND n/a'}, {})
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_na_error_undefok_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND n/a'}, {'FLATFILE': 'NOT FOUND some other error'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: Bestref FAILED:  some other error
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_undefined_defined_undefok_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'UNDEFINED'}, {'FLATFILE': 'x5v1944hl_flat.fits'})
    [UpdateTuple(instrument='COS', filekind='flatfile', old_reference='N/A', new_reference='X5V1944HL_FLAT.FITS')]

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_undefined_na_undefok_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'UNDEFINED'}, {'FLATFILE': 'NOT FOUND n/a'})
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_undefined_undefined_undefok_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'UNDEFINED'}, {'FLATFILE': 'UNDEFINED'})
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_undefined_omitted_undefok_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'UNDEFINED'}, {})
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_undefined_error_undefok_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'UNDEFINED'}, {'FLATFILE': 'NOT FOUND some other error'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: Bestref FAILED:  some other error
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_omitted_defined_undefok_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {}, {'FLATFILE': 'x5v1944hl_flat.fits'})
    [UpdateTuple(instrument='COS', filekind='flatfile', old_reference='N/A', new_reference='X5V1944HL_FLAT.FITS')]

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_omitted_na_undefok_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {}, {'FLATFILE': 'NOT FOUND n/a'})
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_omitted_undefined_undefok_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {}, {'FLATFILE': 'UNDEFINED'})
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_omitted_omitted_undefok_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {}, {})
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_omitted_error_undefok_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {}, {'FLATFILE': 'NOT FOUND some other error'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: Bestref FAILED:  some other error
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_error_defined_undefok_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND some other error'}, {'FLATFILE': 'x5v1944hl_flat.fits'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: Bestref FAILED:  some other error
    [UpdateTuple(instrument='COS', filekind='flatfile', old_reference='NOT FOUND SOME OTHER ERROR', new_reference='X5V1944HL_FLAT.FITS')]

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_error_na_undefok_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND some other error'}, {'FLATFILE': 'NOT FOUND n/a'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: Bestref FAILED:  some other error
    [UpdateTuple(instrument='COS', filekind='flatfile', old_reference='NOT FOUND SOME OTHER ERROR', new_reference='N/A')]

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_error_undefined_undefok_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND some other error'}, {'FLATFILE': 'UNDEFINED'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: Bestref FAILED:  some other error
    [UpdateTuple(instrument='COS', filekind='flatfile', old_reference='NOT FOUND SOME OTHER ERROR', new_reference='N/A')]
    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_error_omitted_undefok_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND some other error'}, {})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: Bestref FAILED:  some other error
    [UpdateTuple(instrument='COS', filekind='flatfile', old_reference='NOT FOUND SOME OTHER ERROR', new_reference='N/A')]

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_error_error_undefok_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND some other error'}, {'FLATFILE': 'NOT FOUND some other error'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: Bestref FAILED:  some other error
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: Bestref FAILED:  some other error
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_defined_defined_undefok_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'x5v1944hl_flat.fits'}, {'FLATFILE': 'x5v1944hl_flat.fits'})
    []

    >>> test_config.cleanup(old_state)
    '''
def dt_compare_bestrefs_defined_defined2_undefok_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'x5v1944hl_flat.fits'}, {'FLATFILE': 'y5v1944hl_flat.fits'})
    [UpdateTuple(instrument='COS', filekind='flatfile', old_reference='X5V1944HL_FLAT.FITS', new_reference='Y5V1944HL_FLAT.FITS')]

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_defined_na_undefok_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'x5v1944hl_flat.fits'}, {'FLATFILE': 'NOT FOUND n/a'})
    [UpdateTuple(instrument='COS', filekind='flatfile', old_reference='X5V1944HL_FLAT.FITS', new_reference='N/A')]

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_defined_undefined_undefok_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'x5v1944hl_flat.fits'}, {'FLATFILE': 'UNDEFINED'})
    [UpdateTuple(instrument='COS', filekind='flatfile', old_reference='X5V1944HL_FLAT.FITS', new_reference='N/A')]

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_defined_omitted_undefok_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'x5v1944hl_flat.fits'}, {})
    [UpdateTuple(instrument='COS', filekind='flatfile', old_reference='X5V1944HL_FLAT.FITS', new_reference='N/A')]

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_defined_error_undefok_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'x5v1944hl_flat.fits'}, {'FLATFILE': 'NOT FOUND some other error'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: Bestref FAILED:  some other error
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_na_defined_undefok_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND n/a'}, {'FLATFILE': 'x5v1944hl_flat.fits'})
    [UpdateTuple(instrument='COS', filekind='flatfile', old_reference='N/A', new_reference='X5V1944HL_FLAT.FITS')]

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_na_na_undefok_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND n/a'}, {'FLATFILE': 'NOT FOUND n/a'})
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_na_undefined_undefok_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND n/a'}, {'FLATFILE': 'UNDEFINED'})
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_na_omitted_undefok_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND n/a'}, {})
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_na_error_undefok_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND n/a'}, {'FLATFILE': 'NOT FOUND some other error'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: Bestref FAILED:  some other error
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_undefined_defined_undefok_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'UNDEFINED'}, {'FLATFILE': 'x5v1944hl_flat.fits'})
    [UpdateTuple(instrument='COS', filekind='flatfile', old_reference='N/A', new_reference='X5V1944HL_FLAT.FITS')]

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_undefined_na_undefok_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'UNDEFINED'}, {'FLATFILE': 'NOT FOUND n/a'})
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_undefined_undefined_undefok_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'UNDEFINED'}, {'FLATFILE': 'UNDEFINED'})
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_undefined_omitted_undefok_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'UNDEFINED'}, {})
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_undefined_error_undefok_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'UNDEFINED'}, {'FLATFILE': 'NOT FOUND some other error'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: Bestref FAILED:  some other error
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_omitted_defined_undefok_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {}, {'FLATFILE': 'x5v1944hl_flat.fits'})
    [UpdateTuple(instrument='COS', filekind='flatfile', old_reference='N/A', new_reference='X5V1944HL_FLAT.FITS')]

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_omitted_na_undefok_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {}, {'FLATFILE': 'NOT FOUND n/a'})
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_omitted_undefined_undefok_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {}, {'FLATFILE': 'UNDEFINED'})
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_omitted_omitted_undefok_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {}, {})
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_omitted_error_undefok_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {}, {'FLATFILE': 'NOT FOUND some other error'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: Bestref FAILED:  some other error
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_error_defined_undefok_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND some other error'}, {'FLATFILE': 'x5v1944hl_flat.fits'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: Bestref FAILED:  some other error
    [UpdateTuple(instrument='COS', filekind='flatfile', old_reference='NOT FOUND SOME OTHER ERROR', new_reference='X5V1944HL_FLAT.FITS')]

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_error_na_undefok_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND some other error'}, {'FLATFILE': 'NOT FOUND n/a'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: Bestref FAILED:  some other error
    [UpdateTuple(instrument='COS', filekind='flatfile', old_reference='NOT FOUND SOME OTHER ERROR', new_reference='N/A')]

    >>> test_config.cleanup(old_state)
    '''

def dt_compare_bestrefs_error_undefined_undefok_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND some other error'}, {'FLATFILE': 'UNDEFINED'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: Bestref FAILED:  some other error
    [UpdateTuple(instrument='COS', filekind='flatfile', old_reference='NOT FOUND SOME OTHER ERROR', new_reference='N/A')]

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_error_omitted_undefok_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND some other error'}, {})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: Bestref FAILED:  some other error
    [UpdateTuple(instrument='COS', filekind='flatfile', old_reference='NOT FOUND SOME OTHER ERROR', new_reference='N/A')]

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_error_error_undefok_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFOK_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND some other error'}, {'FLATFILE': 'NOT FOUND some other error'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: Bestref FAILED:  some other error
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: Bestref FAILED:  some other error
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_defined_defined_undefmatters_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'x5v1944hl_flat.fits'}, {'FLATFILE': 'x5v1944hl_flat.fits'})
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_defined_na_undefmatters_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'x5v1944hl_flat.fits'}, {'FLATFILE': 'NOT FOUND n/a'})
    [UpdateTuple(instrument='COS', filekind='flatfile', old_reference='X5V1944HL_FLAT.FITS', new_reference='N/A')]

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_defined_undefined_undefmatters_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'x5v1944hl_flat.fits'}, {'FLATFILE': 'UNDEFINED'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: No match found: 'UNDEFINED'
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_defined_omitted_undefmatters_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'x5v1944hl_flat.fits'}, {})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: No match found: 'UNDEFINED'
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_defined_error_undefmatters_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'x5v1944hl_flat.fits'}, {'FLATFILE': 'NOT FOUND some other error'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: Bestref FAILED:  some other error
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_na_defined_undefmatters_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND n/a'}, {'FLATFILE': 'x5v1944hl_flat.fits'})
    [UpdateTuple(instrument='COS', filekind='flatfile', old_reference='N/A', new_reference='X5V1944HL_FLAT.FITS')]

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_na_na_undefmatters_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND n/a'}, {'FLATFILE': 'NOT FOUND n/a'})
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_na_undefined_undefmatters_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND n/a'}, {'FLATFILE': 'UNDEFINED'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: No match found: 'UNDEFINED'
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_na_omitted_undefmatters_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND n/a'}, {})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: No match found: 'UNDEFINED'
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_na_error_undefmatters_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND n/a'}, {'FLATFILE': 'NOT FOUND some other error'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: Bestref FAILED:  some other error
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_undefined_defined_undefmatters_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'UNDEFINED'}, {'FLATFILE': 'x5v1944hl_flat.fits'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: No match found: 'UNDEFINED'
    [UpdateTuple(instrument='COS', filekind='flatfile', old_reference='UNDEFINED', new_reference='X5V1944HL_FLAT.FITS')]

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_undefined_na_undefmatters_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'UNDEFINED'}, {'FLATFILE': 'NOT FOUND n/a'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: No match found: 'UNDEFINED'
    [UpdateTuple(instrument='COS', filekind='flatfile', old_reference='UNDEFINED', new_reference='N/A')]

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_undefined_undefined_undefmatters_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'UNDEFINED'}, {'FLATFILE': 'UNDEFINED'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: No match found: 'UNDEFINED'
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: No match found: 'UNDEFINED'
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_undefined_omitted_undefmatters_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'UNDEFINED'}, {})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: No match found: 'UNDEFINED'
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: No match found: 'UNDEFINED'
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_undefined_error_undefmatters_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'UNDEFINED'}, {'FLATFILE': 'NOT FOUND some other error'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: No match found: 'UNDEFINED'
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: Bestref FAILED:  some other error
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_omitted_defined_undefmatters_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {}, {'FLATFILE': 'x5v1944hl_flat.fits'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: No match found: 'UNDEFINED'
    [UpdateTuple(instrument='COS', filekind='flatfile', old_reference='UNDEFINED', new_reference='X5V1944HL_FLAT.FITS')]

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_omitted_na_undefmatters_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {}, {'FLATFILE': 'NOT FOUND n/a'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: No match found: 'UNDEFINED'
    [UpdateTuple(instrument='COS', filekind='flatfile', old_reference='UNDEFINED', new_reference='N/A')]

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_omitted_undefined_undefmatters_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {}, {'FLATFILE': 'UNDEFINED'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: No match found: 'UNDEFINED'
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: No match found: 'UNDEFINED'
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_omitted_omitted_undefmatters_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {}, {})
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_omitted_error_undefmatters_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {}, {'FLATFILE': 'NOT FOUND some other error'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: No match found: 'UNDEFINED'
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: Bestref FAILED:  some other error
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_error_defined_undefmatters_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND some other error'}, {'FLATFILE': 'x5v1944hl_flat.fits'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: Bestref FAILED:  some other error
    [UpdateTuple(instrument='COS', filekind='flatfile', old_reference='NOT FOUND SOME OTHER ERROR', new_reference='X5V1944HL_FLAT.FITS')]

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_error_na_undefmatters_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND some other error'}, {'FLATFILE': 'NOT FOUND n/a'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: Bestref FAILED:  some other error
    [UpdateTuple(instrument='COS', filekind='flatfile', old_reference='NOT FOUND SOME OTHER ERROR', new_reference='N/A')]

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_error_undefined_undefmatters_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND some other error'}, {'FLATFILE': 'UNDEFINED'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: Bestref FAILED:  some other error
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: No match found: 'UNDEFINED'
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_error_omitted_undefmatters_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND some other error'}, {})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: Bestref FAILED:  some other error
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: No match found: 'UNDEFINED'
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_error_error_undefmatters_naok():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAOK.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND some other error'}, {'FLATFILE': 'NOT FOUND some other error'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: Bestref FAILED:  some other error
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: Bestref FAILED:  some other error
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_defined_defined_undefmatters_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'x5v1944hl_flat.fits'}, {'FLATFILE': 'x5v1944hl_flat.fits'})
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_defined_na_undefmatters_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'x5v1944hl_flat.fits'}, {'FLATFILE': 'NOT FOUND n/a'})
    [UpdateTuple(instrument='COS', filekind='flatfile', old_reference='X5V1944HL_FLAT.FITS', new_reference='N/A')]

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_defined_undefined_undefmatters_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'x5v1944hl_flat.fits'}, {'FLATFILE': 'UNDEFINED'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: No match found: 'UNDEFINED'
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_defined_omitted_undefmatters_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'x5v1944hl_flat.fits'}, {})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: No match found: 'UNDEFINED'
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_defined_error_undefmatters_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'x5v1944hl_flat.fits'}, {'FLATFILE': 'NOT FOUND some other error'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: Bestref FAILED:  some other error
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_na_defined_undefmatters_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND n/a'}, {'FLATFILE': 'x5v1944hl_flat.fits'})
    [UpdateTuple(instrument='COS', filekind='flatfile', old_reference='N/A', new_reference='X5V1944HL_FLAT.FITS')]

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_na_na_undefmatters_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND n/a'}, {'FLATFILE': 'NOT FOUND n/a'})
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_na_undefined_undefmatters_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND n/a'}, {'FLATFILE': 'UNDEFINED'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: No match found: 'UNDEFINED'
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_na_omitted_undefmatters_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND n/a'}, {})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: No match found: 'UNDEFINED'
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_na_error_undefmatters_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND n/a'}, {'FLATFILE': 'NOT FOUND some other error'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: Bestref FAILED:  some other error
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_undefined_defined_undefmatters_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'UNDEFINED'}, {'FLATFILE': 'x5v1944hl_flat.fits'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: No match found: 'UNDEFINED'
    [UpdateTuple(instrument='COS', filekind='flatfile', old_reference='UNDEFINED', new_reference='X5V1944HL_FLAT.FITS')]

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_undefined_na_undefmatters_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'UNDEFINED'}, {'FLATFILE': 'NOT FOUND n/a'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: No match found: 'UNDEFINED'
    [UpdateTuple(instrument='COS', filekind='flatfile', old_reference='UNDEFINED', new_reference='N/A')]

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_undefined_undefined_undefmatters_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'UNDEFINED'}, {'FLATFILE': 'UNDEFINED'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: No match found: 'UNDEFINED'
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: No match found: 'UNDEFINED'
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_undefined_omitted_undefmatters_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'UNDEFINED'}, {})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: No match found: 'UNDEFINED'
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: No match found: 'UNDEFINED'
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_undefined_error_undefmatters_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'UNDEFINED'}, {'FLATFILE': 'NOT FOUND some other error'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: No match found: 'UNDEFINED'
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: Bestref FAILED:  some other error
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_omitted_defined_undefmatters_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {}, {'FLATFILE': 'x5v1944hl_flat.fits'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: No match found: 'UNDEFINED'
    [UpdateTuple(instrument='COS', filekind='flatfile', old_reference='UNDEFINED', new_reference='X5V1944HL_FLAT.FITS')]

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_omitted_na_undefmatters_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {}, {'FLATFILE': 'NOT FOUND n/a'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: No match found: 'UNDEFINED'
    [UpdateTuple(instrument='COS', filekind='flatfile', old_reference='UNDEFINED', new_reference='N/A')]

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_omitted_undefined_undefmatters_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {}, {'FLATFILE': 'UNDEFINED'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: No match found: 'UNDEFINED'
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: No match found: 'UNDEFINED'
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_omitted_omitted_undefmatters_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {}, {})
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_omitted_error_undefmatters_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {}, {'FLATFILE': 'NOT FOUND some other error'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: No match found: 'UNDEFINED'
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: Bestref FAILED:  some other error
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_error_defined_undefmatters_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND some other error'}, {'FLATFILE': 'x5v1944hl_flat.fits'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: Bestref FAILED:  some other error
    [UpdateTuple(instrument='COS', filekind='flatfile', old_reference='NOT FOUND SOME OTHER ERROR', new_reference='X5V1944HL_FLAT.FITS')]

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_error_na_undefmatters_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND some other error'}, {'FLATFILE': 'NOT FOUND n/a'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: Bestref FAILED:  some other error
    [UpdateTuple(instrument='COS', filekind='flatfile', old_reference='NOT FOUND SOME OTHER ERROR', new_reference='N/A')]

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_error_undefined_undefmatters_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND some other error'}, {'FLATFILE': 'UNDEFINED'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: Bestref FAILED:  some other error
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: No match found: 'UNDEFINED'
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_error_omitted_undefmatters_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND some other error'}, {})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: Bestref FAILED:  some other error
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: No match found: 'UNDEFINED'
    []

    >>> test_config.cleanup(old_state)
    '''


def dt_compare_bestrefs_error_error_undefmatters_namatters():
    '''
    >>> old_state = test_config.setup()

    >>> SCRIPT_UNDEFMATTERS_NAMATTERS.compare_bestrefs('COS', 'LA9K03C3Q', {'FLATFILE': 'NOT FOUND some other error'}, {'FLATFILE': 'NOT FOUND some other error'})
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  Old: Bestref FAILED:  some other error
    CRDS - ERROR - instrument='COS' type='FLATFILE' data='LA9K03C3Q' ::  New: Bestref FAILED:  some other error
    []

    >>> test_config.cleanup(old_state)
    '''



"""
Tests for crds.misc.check_archive which is nominally used to check the
archive for file availability as an alternative to a true ACK from the
archive when their processing is complete.  Also supplies a function API
used by the server to determine and cache file availability status in the
CRDS catalog.
"""
import sys
import os
import time
import multiprocessing
import tempfile

# ===================================================================

from crds.core import log, config
from crds.misc import check_archive

# ===================================================================

from . import test_config

# ===================================================================

def dt_check_archive_file_api_true():
    """
    >>> old_state = test_config.setup()

    >> check_archive.file_available("hst.pmap")     XXXX DISABLED
    True

    >>> test_config.cleanup(old_state)
    """

def dt_check_archive_file_api_false():
    """
    >>> old_state = test_config.setup()

    >> check_archive.file_available("foo.pmap")     XXXX DISABLED
    CRDS - ERROR -  File 'foo.pmap' failed HTTP HEAD with code = 404 from '.../foo.pmap'
    False

    >>> test_config.cleanup(old_state)
    """

def dt_check_archive_script():
        """
    >>> old_state = test_config.setup()
    >>> check_archive.CheckArchiveScript("crds.misc.check_archive --files foo.map hst.pmap")()
    CRDS - INFO -  Mapping URL: '.../mappings/hst/'
    CRDS - INFO -  Reference URL: '.../references/hst/'
    CRDS - ERROR -  File 'foo.map' failed HTTP HEAD with code = 404 from '.../foo.map'
    CRDS - INFO -        2 files at    ... files-per-second
    CRDS - INFO -  1 errors
    CRDS - INFO -  0 warnings
    CRDS - INFO -  3 infos
    >>> test_config.cleanup(old_state)
    """

# ==================================================================

def main():
    """Run module tests,  for now just doctests only."""
    from crds.tests import test_check_archive, tstmod
    return tstmod(test_check_archive)

if __name__ == "__main__":
    print(main())



import os
import json
import shutil
import datetime

from crds import bestrefs
from crds.bestrefs import BestrefsScript
from crds import assign_bestrefs
from crds.tests import test_config

"""
Bestrefs has a number of command line parameters which make it operate in different modes.

-----------
NEW CONTEXT
-----------

crds.bestrefs always computes best references with respect to a context which can be explicitly specified with the
--new-context parameter.    If no --new-context is specified,  the default operational context is determined by
consulting the CRDS server or looking in the local cache as a fallback.

------------------------
LOOKUP PARAMETER SOURCES
------------------------

The two primary modes for bestrefs involve the source of reference file matching parameters.   Conceptually
lookup parameters are always associated with particular datasets and used to identify the references
required to process those datasets.

The options --files, --datasets, --instruments, and --all determine the source of lookup parameters:

1. To find best references for a list of files do something like this:

    % crds bestrefs --new-context hst.pmap --files j8bt05njq_raw.fits j8bt06o6q_raw.fits j8bt09jcq_raw.fits

the first parameter, hst.pmap,  is the context with respect to which best references are determined.

2. To find best references for a list of catalog dataset ids do something like this:

    % crds bestrefs --new-context hst.pmap --datasets j8bt05njq j8bt06o6q j8bt09jcq

3. To do mass scale testing for all cataloged datasets for a particular instrument(s) do:

    % crds bestrefs --new-context hst.pmap --instruments acs

4. To do mass scale testing for all supported instruments for all cataloged datasets do:

    % crds bestrefs --new-context hst.pmap --all

----------------
COMPARISON MODES
----------------

The --old-context and --compare-source-bestrefs parameters define the best references comparison mode.  Each names
the origin of a set of prior recommendations and implicitly requests a comparison to the recommendations from
the newly computed bestrefs determined by --new-context.

CONTEXT-TO-CONTEXT
..................

--old-context can be used to specify a second context for which bestrefs are dynamically computed; --old-context implies
that a bestrefs comparison will be made with --new-context.

PRIOR SOURCE RECOMMENDATIONS
............................

--compare-source-bestrefs requests that the bestrefs from --new-context be compared to the bestrefs which are
recorded with the lookup parameter data,  either in the file headers of data files,  or in the catalog.   In both
cases the prior best references are recorded static values,  not dynamically computed bestrefs.

------------
UPDATE MODES
------------

Currently there is only one update mode.   When --files are specified as the input source,  --update-bestrefs can
also be specified to update the input data file headers with new bestrefs recommendations.   In this case the data
files are used as both the source of matching parameters and as the destination for best reference recommendations.

------------
OUTPUT MODES
------------

crds.bestrefs supports several output modes for bestrefs and comparison results.

If --print-affected is specified,  crds.bestrefs will print out the name of any file (or dataset id) for which at least one update for
one reference type was recommended.   This is essentially a list of files to be reprocessed with new references.

    % crds bestrefs --new-context hst.pmap --files j8bt05njq_raw.fits j8bt06o6q_raw.fits j8bt09jcq_raw.fits --compare-source-bestrefs --print-affected
    j8bt05njq_raw.fits
    j8bt06o6q_raw.fits
    j8bt09jcq_raw.fits
"""

def dt_bestrefs_3_files():
    """
    Compute simple bestrefs for 3 files:

    >>> old_state = test_config.setup()

    >>> BestrefsScript(argv="bestrefs.py --new-context hst.pmap --files data/j8bt05njq_raw.fits data/j8bt06o6q_raw.fits data/j8bt09jcq_raw.fits")()
    CRDS - INFO - No comparison context or source comparison requested.
    CRDS - INFO - No file header updates requested;  dry run.  Use --update-bestrefs to update FITS headers.
    CRDS - INFO - ===> Processing data/j8bt05njq_raw.fits
    CRDS - INFO - ===> Processing data/j8bt06o6q_raw.fits
    CRDS - INFO - ===> Processing data/j8bt09jcq_raw.fits
    CRDS - INFO - 0 errors
    CRDS - INFO - 0 warnings
    CRDS - INFO - 5 infos
    0

    >>> test_config.cleanup(old_state)
    """

def dt_bestrefs_compare_source_files():
    """
    Compute and print files with at least one reference change:

    >>> old_state = test_config.setup()

    >>> BestrefsScript(argv="bestrefs.py --new-context hst.pmap --files data/j8bt05njq_raw.fits data/j8bt06o6q_raw.fits data/j8bt09jcq_raw.fits --print-affected --compare-source-bestrefs")()
    CRDS - INFO -  No file header updates requested;  dry run.  Use --update-bestrefs to update FITS headers.
    CRDS - INFO -  ===> Processing data/j8bt05njq_raw.fits
    CRDS - INFO -  instrument='ACS' type='ATODTAB' data='data/j8bt05njq_raw.fits' ::  New best reference: 'kcb1734ij_a2d.fits' --> 'n/a' :: Would update.
    CRDS - INFO -  instrument='ACS' type='CRREJTAB' data='data/j8bt05njq_raw.fits' ::  New best reference: 'n4e12510j_crr.fits' --> 'n/a' :: Would update.
    CRDS - INFO -  instrument='ACS' type='IMPHTTAB' data='data/j8bt05njq_raw.fits' ::  New best reference: 'n/a' --> 'w3m1716tj_imp.fits' :: Would update.
    CRDS - INFO -  instrument='ACS' type='NPOLFILE' data='data/j8bt05njq_raw.fits' ::  New best reference: 'n/a' --> 'v9718263j_npl.fits' :: Would update.
    CRDS - INFO -  instrument='ACS' type='SHADFILE' data='data/j8bt05njq_raw.fits' ::  New best reference: 'kcb1734pj_shd.fits' --> 'n/a' :: Would update.
    CRDS - INFO -  ===> Processing data/j8bt06o6q_raw.fits
    CRDS - INFO -  instrument='ACS' type='ATODTAB' data='data/j8bt06o6q_raw.fits' ::  New best reference: 'kcb1734ij_a2d.fits' --> 'n/a' :: Would update.
    CRDS - INFO -  instrument='ACS' type='CRREJTAB' data='data/j8bt06o6q_raw.fits' ::  New best reference: 'n4e12510j_crr.fits' --> 'n/a' :: Would update.
    CRDS - INFO -  instrument='ACS' type='IMPHTTAB' data='data/j8bt06o6q_raw.fits' ::  New best reference: 'n/a' --> 'w3m1716tj_imp.fits' :: Would update.
    CRDS - INFO -  instrument='ACS' type='NPOLFILE' data='data/j8bt06o6q_raw.fits' ::  New best reference: 'n/a' --> 'v9718264j_npl.fits' :: Would update.
    CRDS - INFO -  instrument='ACS' type='SHADFILE' data='data/j8bt06o6q_raw.fits' ::  New best reference: 'kcb1734pj_shd.fits' --> 'n/a' :: Would update.
    CRDS - INFO -  ===> Processing data/j8bt09jcq_raw.fits
    CRDS - INFO -  instrument='ACS' type='ATODTAB' data='data/j8bt09jcq_raw.fits' ::  New best reference: 'kcb1734ij_a2d.fits' --> 'n/a' :: Would update.
    CRDS - INFO -  instrument='ACS' type='IMPHTTAB' data='data/j8bt09jcq_raw.fits' ::  New best reference: 'n/a' --> 'w3m1716tj_imp.fits' :: Would update.
    CRDS - INFO -  instrument='ACS' type='NPOLFILE' data='data/j8bt09jcq_raw.fits' ::  New best reference: 'n/a' --> 'v9718260j_npl.fits' :: Would update.
    CRDS - INFO -  instrument='ACS' type='SHADFILE' data='data/j8bt09jcq_raw.fits' ::  New best reference: 'kcb1734pj_shd.fits' --> 'n/a' :: Would update.
    CRDS - INFO -  Affected products = 3
    data/j8bt05njq_raw.fits
    data/j8bt06o6q_raw.fits
    data/j8bt09jcq_raw.fits
    CRDS - INFO -  0 errors
    CRDS - INFO -  0 warnings
    CRDS - INFO -  19 infos
    0

    >>> test_config.cleanup(old_state)
    """

def dt_bestrefs_3_files_default_context_from_server():
    """
    Compute simple bestrefs for 3 files using the default context from the server:

    >>> old_state = test_config.setup()

    >>> BestrefsScript(argv="bestrefs.py --new-context=hst.pmap --files data/j8bt05njq_raw.fits data/j8bt06o6q_raw.fits data/j8bt09jcq_raw.fits")()
    CRDS - INFO - No comparison context or source comparison requested.
    CRDS - INFO - No file header updates requested;  dry run.  Use --update-bestrefs to update FITS headers.
    CRDS - INFO - ===> Processing data/j8bt05njq_raw.fits
    CRDS - INFO - ===> Processing data/j8bt06o6q_raw.fits
    CRDS - INFO - ===> Processing data/j8bt09jcq_raw.fits
    CRDS - INFO - 0 errors
    CRDS - INFO - 0 warnings
    CRDS - INFO - 5 infos
    0

    >>> test_config.cleanup(old_state)
    """

def dt_bestrefs_broken_dataset_file():
    """
    Same + one broken file to test shell error status

    >>> old_state = test_config.setup()

    >>> BestrefsScript(argv="bestrefs.py --new-context hst.pmap --files data/j8bt05njq_raw.fits data/j8bt05njq_raw_broke.fits data/j8bt06o6q_raw.fits data/j8bt09jcq_raw.fits")()
    CRDS - INFO - No comparison context or source comparison requested.
    CRDS - INFO - No file header updates requested;  dry run.  Use --update-bestrefs to update FITS headers.
    CRDS - INFO - ===> Processing data/j8bt05njq_raw.fits
    CRDS - INFO - ===> Processing data/j8bt05njq_raw_broke.fits
    CRDS - ERROR - instrument='ACS' type='BIASFILE' data='data/j8bt05njq_raw_broke.fits' ::  New: Bestref FAILED:   parameter='CCDAMP' value='FOOBAR' is not in ['A', 'ABCD', 'AC', 'AD', 'B', 'BC', 'BD', 'C', 'D']
    CRDS - INFO - ===> Processing data/j8bt06o6q_raw.fits
    CRDS - INFO - ===> Processing data/j8bt09jcq_raw.fits
    CRDS - INFO - 1 errors
    CRDS - INFO - 0 warnings
    CRDS - INFO - 6 infos
    1

    >>> test_config.cleanup(old_state)
    """

def dt_bestrefs_broken_cache_and_server():
    """


    >>> old_state = test_config.setup(cache="/nowhere", url="https://server-is-out-of-town")

    >> BestrefsScript(argv="bestrefs.py --new-context hst.pmap --files data/j8bt05njq_raw.fits")()
    CRDS - ERROR - (FATAL) CRDS server connection and cache load FAILED.  Cannot continue.  See https://hst-crds.stsci.edu or https://jwst-crds.stsci.edu for more information on configuring CRDS.
    Traceback (most recent call last):
    ...
    SystemExit: 1

>>> test_config.cleanup(old_state)
    """

def dt_bestrefs_catalog_dataset():
    """
    Compute simple bestrefs for 1 catalog datasets using hst.pmap:

    >>> old_state = test_config.setup()

    >>> BestrefsScript(argv="bestrefs.py --new-context hst.pmap  --datasets LB6M01030")() # doctest: +ELLIPSIS
    CRDS - INFO - Dumping dataset parameters from CRDS server at '...' for ['LB6M01030']
    CRDS - INFO - Dumped 1 of 1 datasets from CRDS server at '...'
    CRDS - INFO - Computing bestrefs for datasets ['LB6M01030']
    CRDS - INFO - No comparison context or source comparison requested.
    CRDS - INFO - 0 errors
    CRDS - INFO - 0 warnings
    CRDS - INFO - 4 infos
    0

    >>> test_config.cleanup(old_state)

    MAINTENANCE NOTE:  the preceding test is currently an expected error case pending the delivery of a modified
    WFC3 FLSHFILE rmap located at crds/hst/prototypes/wfc3/hst_wfc3_flshfile_0251.rmap.  Once the modified rmap
    is delivered to operations,  the above new-context should be changed to the new OPS context.  After that point,
    all mirrors of OPS to DEV should work without the exected errors due to FLASHCUR=='UNDEFINED'.   The only changes
    in the modified rmap should be header changes,  nominally the rmap_relevance expression;  additional changes
    may reflect new flshfile submissions which happened after the prototype rmap was created.
    """

def dt_bestrefs_context_to_context():
    """
    Compute comparison bestrefs between two contexts:

    >>> old_state = test_config.setup()

    >>> BestrefsScript(argv="bestrefs.py --new-context data/hst_0001.pmap  --old-context hst.pmap --files data/j8bt05njq_raw.fits data/j8bt06o6q_raw.fits data/j8bt09jcq_raw.fits")()
    CRDS - INFO - No file header updates requested;  dry run.  Use --update-bestrefs to update FITS headers.
    CRDS - INFO - ===> Processing data/j8bt05njq_raw.fits
    CRDS - INFO - ===> Processing data/j8bt06o6q_raw.fits
    CRDS - INFO - ===> Processing data/j8bt09jcq_raw.fits
    CRDS - INFO - 0 errors
    CRDS - INFO - 0 warnings
    CRDS - INFO - 4 infos
    0

    >>> test_config.cleanup(old_state)
    """

def dt_bestrefs_all_instruments_hst():
    """
    Compute comparison bestrefs between two contexts:

    >>> old_state = test_config.setup()
    >>> script = BestrefsScript(argv="bestrefs.py --new-context data/hst_0001.pmap  --old-context hst.pmap --all-instruments") # doctest: +ELLIPSIS
    >>> script.complex_init()
    CRDS - INFO -  Computing bestrefs for db datasets for ['acs', 'cos', 'nicmos', 'stis', 'wfc3', 'wfpc2']
    CRDS - INFO -  Dumping dataset parameters for 'acs' from CRDS server at '...'
    CRDS - INFO -  Downloaded  ... dataset ids for 'acs' since None
    CRDS - INFO -  Dumping dataset parameters for 'cos' from CRDS server at '...'
    CRDS - INFO -  Downloaded  ... dataset ids for 'cos' since None
    CRDS - INFO -  Dumping dataset parameters for 'nicmos' from CRDS server at '...'
    CRDS - INFO -  Downloaded  ... dataset ids for 'nicmos' since None
    CRDS - INFO -  Dumping dataset parameters for 'stis' from CRDS server at '...'
    CRDS - INFO -  Downloaded  ... dataset ids for 'stis' since None
    CRDS - INFO -  Dumping dataset parameters for 'wfc3' from CRDS server at '...'
    CRDS - INFO -  Downloaded  ... dataset ids for 'wfc3' since None
    CRDS - INFO -  Dumping dataset parameters for 'wfpc2' from CRDS server at '...'
    CRDS - INFO -  Downloaded  ... dataset ids for 'wfpc2' since None
    True
    >>> test_config.cleanup(old_state)
    """

def dt_bestrefs_datasets_since_auto_hst():
    """
    Compute comparison bestrefs between two contexts:

    >>> old_state = test_config.setup()
    >>> script = BestrefsScript(argv="bestrefs.py --old-context hst.pmap --new-context data/hst_0001.pmap --diffs-only --datasets-since=auto") # doctest: +ELLIPSIS
    >>> script.complex_init()
    CRDS - INFO -  Mapping differences from 'hst.pmap' --> 'data/hst_0001.pmap' affect:
     {'acs': ['biasfile']}
    CRDS - INFO -  Possibly affected --datasets-since dates determined by 'hst.pmap' --> 'data/hst_0001.pmap' are:
     {'acs': '1992-01-02 00:00:00'}
    CRDS - INFO -  Computing bestrefs for db datasets for ['acs']
    CRDS - INFO -  Dumping dataset parameters for 'acs' from CRDS server at '...' since '1992-01-02 00:00:00'
    CRDS - INFO -  Downloaded  ... dataset ids for 'acs' since '1992-01-02 00:00:00'
    True
    >>> test_config.cleanup(old_state)
    """

def dt_bestrefs_dataset_drop_ids():
    """
    >>> old_state = test_config.setup()
    >>> script = BestrefsScript(argv="bestrefs.py --new-context data/hst_0001.pmap  --old-context hst.pmap --load-pickle data/test_cos.json --drop-ids LCE31SW6Q:LCE31SW6Q")
    >>> script.complex_init()
    CRDS - INFO -  Loading file 'data/test_cos.json'
    CRDS - INFO -  Loaded 1 datasets from file 'data/test_cos.json' completely replacing existing headers.
    True
    >>> test_config.cleanup(old_state)
    """

def dt_bestrefs_dataset_only_ids():
    """
    >>> old_state = test_config.setup()
    >>> script = BestrefsScript(argv="bestrefs.py --new-context data/hst_0001.pmap  --old-context hst.pmap --load-pickle data/test_cos.json --only-ids LPPPPPP6Q:LCE31SW6Q")
    >>> script.complex_init()
    CRDS - INFO -  Loading file 'data/test_cos.json'
    CRDS - INFO -  Loaded 1 datasets from file 'data/test_cos.json' completely replacing existing headers.
    True
    >>> test_config.cleanup(old_state)
    """

def dt_bestrefs_compare_source_canary():
    """
    >>> old_state = test_config.setup()
    >>> BestrefsScript("crds.bestrefs --new-context hst_0551.pmap --compare-source --load-pickles data/canary.json --differences-are-errors")()  # doctest: +ELLIPSIS
    CRDS - INFO -  Loading file 'data/canary.json'
    CRDS - INFO -  Loaded 1 datasets from file 'data/canary.json' completely replacing existing headers.
    CRDS - ERROR -  instrument='COS' type='BPIXTAB' data='LA7803FIQ' ::  Comparison difference: 'bar.fits' --> 'yae1249sl_bpix.fits' :: Would update.
    CRDS - ERROR -  instrument='COS' type='XWLKFILE' data='LA7803FIQ' ::  Comparison difference: 'foo.fits' --> '14o2013ql_xwalk.fits' :: Would update.
    CRDS - INFO -  2 errors
    CRDS - INFO -  0 warnings
    CRDS - INFO -  2 infos
    2
    >>> test_config.cleanup(old_state)
    """

def dt_bestrefs_multiple_updates_with_error():
    """
    >>> old_state = test_config.setup()
    >>> BestrefsScript("crds.bestrefs --old-context hst_0628.pmap --new-context hst_0633.pmap --print-affected --load-pickle data/bestrefs_new_error.json --verbosity=30")() # doctest: +ELLIPSIS
    CRDS - DEBUG -  Command: ['crds.bestrefs', '--old-context', 'hst_0628.pmap', '--new-context', 'hst_0633.pmap', '--print-affected', '--load-pickle', 'data/bestrefs_new_error.json', '--verbosity=30']
    CRDS - DEBUG -  Using explicit new context 'hst_0633.pmap' for computing updated best references.
    CRDS - DEBUG -  Using explicit old context 'hst_0628.pmap'
    CRDS - INFO -  Loading file 'data/bestrefs_new_error.json'
    CRDS - INFO -  Loaded 1 datasets from file 'data/bestrefs_new_error.json' completely replacing existing headers.
    CRDS - DEBUG -  ===> Processing JA9553M3Q:JA9553M3Q
    CRDS - DEBUG -  instrument='ACS' type='ATODTAB' data='JA9553M3Q:JA9553M3Q' ::  Lookup MATCHES: 'n/a' :: No update.
    CRDS - ERROR -  instrument='ACS' type='BIASFILE' data='JA9553M3Q' ::  New: Bestref FAILED:   parameter='CCDGAIN' value='1.4' is not in ['1.0', '2.0', '4.0', '8.0']
    CRDS - DEBUG -  instrument='ACS' type='BPIXTAB' data='JA9553M3Q:JA9553M3Q' ::  Lookup MATCHES: 't3n1116nj_bpx.fits' :: No update.
    CRDS - DEBUG -  instrument='ACS' type='CCDTAB' data='JA9553M3Q:JA9553M3Q' ::  Lookup MATCHES: 'uc82140bj_ccd.fits' :: No update.
    CRDS - DEBUG -  instrument='ACS' type='CFLTFILE' data='JA9553M3Q:JA9553M3Q' ::  Lookup MATCHES: 'n/a' :: No update.
    CRDS - DEBUG -  instrument='ACS' type='CRREJTAB' data='JA9553M3Q:JA9553M3Q' ::  Lookup MATCHES: 'n/a' :: No update.
    CRDS - DEBUG -  instrument='ACS' type='D2IMFILE' data='JA9553M3Q:JA9553M3Q' ::  Lookup MATCHES: 'n/a' :: No update.
    CRDS - INFO -  instrument='ACS' type='DARKFILE' data='JA9553M3Q:JA9553M3Q' ::  New best reference: '1ag2019jj_drk.fits' --> '25815071j_drk.fits' :: Would update.
    CRDS - DEBUG -  instrument='ACS' type='DGEOFILE' data='JA9553M3Q:JA9553M3Q' ::  Lookup MATCHES: 'qbu16423j_dxy.fits' :: No update.
    CRDS - INFO -  instrument='ACS' type='DRKCFILE' data='JA9553M3Q:JA9553M3Q' ::  New best reference: '1ag20119j_dkc.fits' --> '2581508ij_dkc.fits' :: Would update.
    CRDS - DEBUG -  instrument='ACS' type='FLSHFILE' data='JA9553M3Q:JA9553M3Q' ::  Lookup MATCHES: 'n/a' :: No update.
    CRDS - DEBUG -  instrument='ACS' type='IDCTAB' data='JA9553M3Q:JA9553M3Q' ::  Lookup MATCHES: '0461802dj_idc.fits' :: No update.
    CRDS - DEBUG -  instrument='ACS' type='IMPHTTAB' data='JA9553M3Q:JA9553M3Q' ::  Lookup MATCHES: 'n/a' :: No update.
    CRDS - DEBUG -  instrument='ACS' type='MDRIZTAB' data='JA9553M3Q:JA9553M3Q' ::  Lookup MATCHES: 'n/a' :: No update.
    CRDS - DEBUG -  instrument='ACS' type='MLINTAB' data='JA9553M3Q:JA9553M3Q' ::  Lookup MATCHES: 'n/a' :: No update.
    CRDS - DEBUG -  instrument='ACS' type='NPOLFILE' data='JA9553M3Q:JA9553M3Q' ::  Lookup MATCHES: 'n/a' :: No update.
    CRDS - DEBUG -  instrument='ACS' type='OSCNTAB' data='JA9553M3Q:JA9553M3Q' ::  Lookup MATCHES: '17717071j_osc.fits' :: No update.
    CRDS - DEBUG -  instrument='ACS' type='PCTETAB' data='JA9553M3Q:JA9553M3Q' ::  Lookup MATCHES: '19i16323j_cte.fits' :: No update.
    CRDS - DEBUG -  instrument='ACS' type='PFLTFILE' data='JA9553M3Q:JA9553M3Q' ::  Lookup MATCHES: 'n/a' :: No update.
    CRDS - DEBUG -  instrument='ACS' type='SHADFILE' data='JA9553M3Q:JA9553M3Q' ::  Lookup MATCHES: 'n/a' :: No update.
    CRDS - DEBUG -  instrument='ACS' type='SNKCFILE' data='JA9553M3Q:JA9553M3Q' ::  Lookup MATCHES: 'n/a' :: No update.
    CRDS - DEBUG -  instrument='ACS' type='SPOTTAB' data='JA9553M3Q:JA9553M3Q' ::  Lookup MATCHES: 'n/a' :: No update.
    CRDS - INFO -  Affected products = 0
    CRDS - DEBUG -  1 sources processed
    CRDS - DEBUG -  0 source updates
    CRDS - INFO -  1 errors
    CRDS - INFO -  0 warnings
    CRDS - INFO -  5 infos
    1
    >>> test_config.cleanup(old_state)
    """

def dt_test_cleanpath():
    """
    Removes prefixes added to reference files prior to writing into FITS headers.

    HST IRAF-style env path prefix

    >>> bestrefs.cleanpath("jref$foo.fits")
    'foo.fits'

    JWST URI-style prefix

    >>> bestrefs.cleanpath('crds://foo.fits')
    'foo.fits'

    >>> bestrefs.cleanpath('foo.fits')
    'foo.fits'

    >>> bestrefs.cleanpath('something/foo.fits')
    'something/foo.fits'
    """

def dt_test_hst_tobs_header_to_reftypes():
    """
    >>> from crds.hst.locate import header_to_reftypes
    >>> header_to_reftypes({}, "hst.pmap")
    []
    >>> from crds.tobs.locate import header_to_reftypes
    >>> header_to_reftypes({}, "tobs.pmap")
    []
    """

def dt_test_jwst_header_to_reftypes():
    """
    >>> from crds.jwst.locate import header_to_reftypes

    >>> header_to_reftypes({"EXP_TYPE":"MIR_DARK", "CAL_VER": "0.7.0"}, "jwst_0301.pmap")
    ['ipc', 'linearity', 'mask', 'refpix', 'rscd', 'saturation', 'superbias']

    >>> header_to_reftypes({"EXP_TYPE":"NIR_IMAGE", "CAL_VER": "0.7.0"}, "jwst_0301.pmap")
    []

    # Traceback (most recent call last):
    # ...
    # RuntimeError: Unhandled EXP_TYPE 'NIR_IMAGE'

    >>> header_to_reftypes({"EXP_TYPE":"MIR_IMAGE", "CAL_VER": "0.7.0"}, "jwst_0301.pmap")
    ['area', 'camera', 'collimator', 'dark', 'disperser', 'distortion', 'filteroffset', 'fore', 'fpa', 'gain', 'ifufore', 'ifupost', 'ifuslicer', 'ipc', 'linearity', 'mask', 'msa', 'ote', 'photom', 'readnoise', 'refpix', 'regions', 'rscd', 'saturation', 'specwcs', 'superbias', 'v2v3', 'wavelengthrange']
    """

class TestBestrefs(test_config.CRDSTestCase):

    script_class = BestrefsScript
    # server_url = "https://hst-crds-dev.stsci.edu"
    cache = test_config.CRDS_TESTING_CACHE

    def get_10_days_ago(self):
        now = datetime.datetime.now()
        now -= datetime.timedelta(days=10)
        return now.isoformat().split("T")[0]

    def test_bestrefs_affected_datasets(self):
        self.run_script(f"crds.bestrefs --affected-datasets --old-context hst_0943.pmap --new-context hst_0947.pmap "
                        f"--datasets-since {self.get_10_days_ago()}",
                        expected_errs=0)

    def test_bestrefs_from_pickle(self):
        self.run_script("crds.bestrefs --new-context hst_0315.pmap --load-pickle data/test_cos.pkl --stats --print-affected-details",
                        expected_errs=0)

    def test_bestrefs_to_pickle(self):
        self.run_script("crds.bestrefs --datasets LA9K03C3Q:LA9K03C3Q LA9K03C5Q:LA9K03C5Q LA9K03C7Q:LA9K03C7Q "
                        "--new-context hst_0315.pmap --save-pickle test_cos.pkl --stats",
                        expected_errs=0)
        os.remove("test_cos.pkl")

    def test_bestrefs_from_json(self):
        self.run_script("crds.bestrefs --new-context hst_0315.pmap --load-pickle data/test_cos.json --stats",
                        expected_errs=1)

    def test_bestrefs_to_json(self):
        self.run_script(f"crds.bestrefs --instrument cos --new-context hst_0315.pmap --save-pickle test_cos.json "
                        f"--datasets-since {self.get_10_days_ago()}", expected_errs=None)
        os.remove("test_cos.json")

    def test_bestrefs_at_file(self):
        self.run_script("crds.bestrefs --files @data/bestrefs_file_list  --new-context hst_0315.pmap --stats",
                        expected_errs=0)

    def test_bestrefs_remote(self):
        self.run_script("crds.bestrefs --files @data/bestrefs_file_list  --new-context hst_0315.pmap --remote --stats",
                        expected_errs=0)

    def test_bestrefs_new_references(self):
        self.run_script("crds.bestrefs --files @data/bestrefs_file_list  --new-context hst_0315.pmap --print-new-references --stats",
                        expected_errs=0)

    def test_bestrefs_default_new_context(self):
        self.run_script("crds.bestrefs --files @data/bestrefs_file_list  --stats",
                        expected_errs=0)

    def test_bestrefs_update_file_headers(self):
        shutil.copy("data/j8bt06o6q_raw.fits", "j8bt06o6q_raw.fits")
        self.run_script("crds.bestrefs --files ./j8bt06o6q_raw.fits --new-context hst_0315.pmap --update-bestrefs",
                       expected_errs=0)
        os.remove("j8bt06o6q_raw.fits")

    def test_bestrefs_update_bestrefs(self):
        # """update_bestrefs modifies dataset file headers"""
        shutil.copy("data/j8bt06o6q_raw.fits", "j8bt06o6q_raw.fits")
        self.run_script("crds.bestrefs --files ./j8bt06o6q_raw.fits --new-context hst_0315.pmap --update-bestrefs",
                       expected_errs=0)
        os.remove("j8bt06o6q_raw.fits")

    def test_bestrefs_bad_sources(self):
        with self.assertRaises(AssertionError):
            self.run_script("crds.bestrefs --all-instruments --instrument cos --new-context hst_0315.pmap",
                            expected_errs=1)

    def test_bestrefs_update_headers(self):
        # """update_headers updates original headers from a pickle saving a new pickle withn orginal + overrides."""
        self.run_script("crds.bestrefs --new-context hst_0315.pmap --datasets LCE31SW6Q:LCE31SW6Q --load-pickle data/test_cos_update.json "
                        " --save-pickle ./test_cos_combined.json --update-bestrefs --update-pickle", expected_errs=1)
        with open("./test_cos_combined.json") as pfile:
            header = json.load(pfile)
        header = header["LCE31SW6Q:LCE31SW6Q"]
        badttab = header["BADTTAB"]
        self.assertEqual(badttab, "N/A")
        gsagtab = header["GSAGTAB"]
        self.assertEqual(gsagtab, "X6L1439EL_GSAG.FITS")
        flatfile = header["FLATFILE"]
        self.assertEqual(flatfile, "N/A")
        os.remove("./test_cos_combined.json")

    def test_assign_bestrefs(self):
        test_copy = "cos_N8XTZCAWQ_updated.fits"
        shutil.copy("data/cos_N8XTZCAWQ.fits", test_copy)

        errors = assign_bestrefs([test_copy], context="hst_0500.pmap")
        self.assertEqual(errors, 0)

        os.remove(test_copy)


# ==================================================================================

def main():
    """Run module tests,  for now just doctests only."""
    import unittest
    suite = unittest.TestLoader().loadTestsFromTestCase(TestBestrefs)
    unittest.TextTestRunner().run(suite)

    from crds.tests import test_bestrefs, tstmod
    return tstmod(test_bestrefs)

if __name__ == "__main__":
    print(main())



"""Higher level mapping based tests for selectors not covered by hst.
"""
import sys
import os
import unittest

import crds
from crds.core import rmap, log
from crds.tests import test_config

# =============================================================================

class TobsTestCase(test_config.CRDSTestCase):
    cache = test_config.CRDS_TESTING_CACHE
    clear_existing = False
    server_url = "https://tobs-serverless-mode.stsci.edu"

class Test_00_Selectors(TobsTestCase):

    def setUp(self):
        super(Test_00_Selectors, self).setUp()
        self.rmap = rmap.get_cached_mapping("tobs_tinstr_tfilekind.rmap")

    def _selector_testcase(self, case, parameter, result):
        header = {
            "TEST_CASE": case,
            "PARAMETER": parameter,
        }
        bestref = self.rmap.get_best_ref(header)
        self.assertEqual(bestref, result)

    def test_use_after_bad_datetime(self):
        header = { "TEST_CASE":"USE_AFTER", "PARAMETER": '4.5' }
        bestref = self.rmap.get_best_ref(header)
        assert bestref.startswith("NOT FOUND UseAfter Invalid date/time format")

    def test_use_after_no_time(self):
        self._selector_testcase('USE_AFTER', '2005-12-20', 'o9f15549j_bia.fits')

    def test_use_after_nominal(self):
        self._selector_testcase('USE_AFTER', '2005-12-20 12:00:00', 'o9f15549j_bia.fits')

    def test_use_after_tuple(self):
        self._selector_testcase('USE_AFTER',  '2004-04-25 21:31:01', ('foo_bia.fits', 'bar_bia.fits'))

    def test_use_after_dict(self):
        self._selector_testcase('USE_AFTER',  '2004-04-25 21:31:02', {'foo':'foo_bia1.fits', 'bar':'bar_bia2.fits'})

    def test_use_after_missing_parameter(self):
        header = { "TEST_CASE": "USE_AFTER" }  # no PARAMETER
        bestref = self.rmap.get_best_ref(header)
        assert bestref.startswith("NOT FOUND UseAfter required lookup parameter 'PARAMETER' is undefined.")

    def test_select_version1(self):
        self._selector_testcase('SELECT_VERSION', '4.5', 'cref_flatfield_73.fits')
    def test_select_version2(self):
        self._selector_testcase('SELECT_VERSION', '5', 'cref_flatfield_123.fits')
    def test_select_version3(self):
        self._selector_testcase('SELECT_VERSION', '6', 'cref_flatfield_123.fits')
    def test_select_version4(self):
        self._selector_testcase('SELECT_VERSION', '2.0', 'cref_flatfield_65.fits')

    def test_closest_time1(self):
        self._selector_testcase('CLOSEST_TIME', '2016-05-05', 'cref_flatfield_123.fits')
    def test_closest_time2(self):
        self._selector_testcase('CLOSEST_TIME', '2016-04-24', 'cref_flatfield_123.fits')
    def test_closest_time3(self):
        self._selector_testcase('CLOSEST_TIME', '2018-02-02', 'cref_flatfield_222.fits')
    def test_closest_time4(self):
        self._selector_testcase('CLOSEST_TIME', '2019-03-01', 'cref_flatfield_123.fits')
    def test_closest_time5(self):
        self._selector_testcase('CLOSEST_TIME', '2016-04-15', 'cref_flatfield_123.fits')
    def test_closest_time6(self):
        self._selector_testcase('CLOSEST_TIME', '2016-04-16', 'cref_flatfield_123.fits')

    def test_bracket1(self):
        self._selector_testcase('BRACKET', '1.25',
            ('cref_flatfield_120.fits', 'cref_flatfield_124.fits'))
    def test_bracket2(self):
        self._selector_testcase('BRACKET', '1.2',
            ('cref_flatfield_120.fits', 'cref_flatfield_120.fits'))
    def test_bracket3(self):
        self._selector_testcase('BRACKET', '1.5',
            ('cref_flatfield_124.fits', 'cref_flatfield_124.fits'))
    def test_bracket4(self):
        self._selector_testcase('BRACKET', '5.0',
            ('cref_flatfield_137.fits', 'cref_flatfield_137.fits'))
    def test_bracket5(self):
        self._selector_testcase('BRACKET', '1.0',
            ('cref_flatfield_120.fits', 'cref_flatfield_120.fits'))
    def test_bracket6(self):
        self._selector_testcase('BRACKET', '6.0',
            ('cref_flatfield_137.fits', 'cref_flatfield_137.fits'))

    def test_geometrically_nearest1(self):
        self._selector_testcase("GEOMETRICALLY_NEAREST", '1.0', 'cref_flatfield_120.fits')
    def test_geometrically_nearest2(self):
        self._selector_testcase("GEOMETRICALLY_NEAREST", '1.2', 'cref_flatfield_120.fits')
    def test_geometrically_nearest3(self):
        self._selector_testcase("GEOMETRICALLY_NEAREST", '1.25', 'cref_flatfield_120.fits')
    def test_geometrically_nearest4(self):
        self._selector_testcase("GEOMETRICALLY_NEAREST", '1.4', 'cref_flatfield_124.fits')
    def test_geometrically_nearest5(self):
        self._selector_testcase("GEOMETRICALLY_NEAREST", '3.25', 'cref_flatfield_124.fits')
    def test_geometrically_nearest6(self):
        self._selector_testcase("GEOMETRICALLY_NEAREST", '3.26', 'cref_flatfield_137.fits')
    def test_geometrically_nearest7(self):
        self._selector_testcase("GEOMETRICALLY_NEAREST", '5.0', 'cref_flatfield_137.fits')
    def test_geometrically_nearest8(self):
        self._selector_testcase("GEOMETRICALLY_NEAREST", '5.1', 'cref_flatfield_137.fits')

    def test_reference_names(self):
        assert self.rmap.reference_names() == ['bar_bia.fits', 'bar_bia2.fits', 'cref_flatfield_120.fits',
                                               'cref_flatfield_123.fits', 'cref_flatfield_124.fits',
                                               'cref_flatfield_137.fits', 'cref_flatfield_222.fits',
                                               'cref_flatfield_65.fits', 'cref_flatfield_73.fits',
                                               'foo_bia.fits', 'foo_bia1.fits', 'nal1503ij_bia.fits',
                                               'o3913216j_bia.fits', 'o5d10135j_bia.fits', 'o9f15549j_bia.fits',
                                               'o9s16388j_bia.fits', 'o9t1525sj_bia.fits']

# =============================================================================

class Test_01_Insert(TobsTestCase):
    """Tests for checking automatic rmap update logic for adding new references."""

    def setUp(self):
        # Note:  load_mapping must deliver a unique copy of the specified rmap
        super(Test_01_Insert, self).setUp()
        self.rmap = rmap.load_mapping("tobs_tinstr_tfilekind.rmap")
        self.original = rmap.load_mapping("tobs_tinstr_tfilekind.rmap")

    def class_name(self, selector_name):
        return "".join([x.capitalize() for x in selector_name.split("_")])

    def set_classes(self, classes):
        self.rmap.selector._rmap_header["classes"] = classes

    def terminal_insert(self, selector_name, param, value):
        """Check the bottom level insert functionality."""
        header = {
                  "TEST_CASE" : selector_name,
                  "PARAMETER" : param,
        }
        inner_class = self.class_name(selector_name)
        self.set_classes(("Match", inner_class))
        result = self.rmap.insert(header, value)
        diffs = self.rmap.difference(result)
        log.verbose("diffs:", diffs)
        assert len(diffs) == 1, "Fewer/more differences than expected"
        assert diffs[0][0] == ('tobs_tinstr_tfilekind.rmap', 'tobs_tinstr_tfilekind.rmap'), "unexpected file names in diff"
        assert diffs[0][1] == (selector_name,), "unexpected match case in diff"
        assert diffs[0][2] == (str(param),), "unexpected parameter value in diff"
        assert diffs[0][3] == "added terminal " + repr(value), "diff is not an addition " + repr(diffs[0])

    def terminal_replace(self, selector_name, param, value):
        """Check the bottom level replace functionality."""
        header = {
                  "TEST_CASE" : selector_name,
                  "PARAMETER" : param,
        }
        inner_class = self.class_name(selector_name)
        self.set_classes(("Match", inner_class))
        result = self.rmap.insert(header, value)
        diffs = self.rmap.difference(result)
        log.verbose("diffs:", diffs)
        assert len(diffs) == 1, "Fewer/more differences than expected"
        assert diffs[0][0] == ('tobs_tinstr_tfilekind.rmap', 'tobs_tinstr_tfilekind.rmap'), "unexpected file names in diff"
        assert diffs[0][1] == (selector_name,), "unexpected match case in diff"
        assert diffs[0][2] == (str(param),), "unexpected parameter value in diff"
        assert "replaced" in diffs[0][3], "diff is not a replacement " + repr(diffs[0])
        assert diffs[0][3].endswith(repr(value))

    def test_useafter_insert_before(self):
        self.terminal_insert("USE_AFTER", '2003-09-25 01:28:00', 'foo.fits')

    def test_useafter_replace_before(self):
        self.terminal_replace("USE_AFTER", '2003-09-26 01:28:00', 'foo.fits')

    def test_useafter_insert_mid(self):
        self.terminal_insert("USE_AFTER", '2004-06-18 04:36:01', 'foo.fits')

    def test_useafter_replace_mid(self):
        self.terminal_replace("USE_AFTER", '2004-06-18 04:36:00', 'foo.fits')

    def test_useafter_insert_after(self):
        self.terminal_insert("USE_AFTER", '2004-07-14 16:52:01', 'foo.fits')

    def test_useafter_replace_after(self):
        self.terminal_replace("USE_AFTER", '2004-07-14 16:52:00', 'foo.fits')

    """
           '<3.1':    'cref_flatfield_65.fits',
           '<5':      'cref_flatfield_73.fits',
           'default': 'cref_flatfield_123.fits',
    """
    def test_select_version_insert_before(self):
        self.terminal_insert("SELECT_VERSION", '<3.0', 'foo.fits')

    def test_select_version_insert_mid(self):
        self.terminal_insert("SELECT_VERSION", '<4', 'foo.fits')

    # There's *nothing* after default,  so no insert possible.
    #    def test_select_version_insert_after(self):
    #        self.terminal_insert("SELECT_VERSION", 'default', 'foo.fits')

    def test_select_version_replace_before(self):
        self.terminal_replace("SELECT_VERSION", '<3.1', 'foo.fits')

    def test_select_version_replace_mid(self):
        self.terminal_replace("SELECT_VERSION", '<5', 'foo.fits')

    def test_select_version_replace_after(self):
        self.terminal_replace("SELECT_VERSION", 'default', 'foo.fits')

    """
            '2017-04-24': "cref_flatfield_123.fits",
            '2018-02-01':  "cref_flatfield_222.fits",
            '2019-04-15': "cref_flatfield_123.fits",
    """
    def test_closest_time_insert_before(self):
         self.terminal_insert("CLOSEST_TIME", '2017-04-20 00:00:00', 'foo.fits')
    def test_closest_time_insert_mid(self):
         self.terminal_insert("CLOSEST_TIME", '2018-01-20 00:57', 'foo.fits')
    def test_closest_time_insert_after(self):
         self.terminal_insert("CLOSEST_TIME", '2020-04-20 00:00:00', 'foo.fits')

    def test_closest_time_replace_before(self):
         self.terminal_replace("CLOSEST_TIME", '2017-04-24', 'foo.fits')
    def test_closest_time_replace_mid(self):
         self.terminal_replace("CLOSEST_TIME", '2018-02-01', 'foo.fits')
    def test_closest_time_replace_after(self):
         self.terminal_replace("CLOSEST_TIME", '2019-04-15', 'foo.fits')

    """
            1.2: "cref_flatfield_120.fits",
            1.5: "cref_flatfield_124.fits",
            5.0: "cref_flatfield_137.fits",
    """
    def test_bracket_insert_before(self):
         self.terminal_insert("BRACKET", 1.0, 'foo.fits')
    def test_bracket_insert_mid(self):
         self.terminal_insert("BRACKET", 1.3, 'foo.fits')
    def test_bracket_insert_after(self):
         self.terminal_insert("BRACKET", 5.1, 'foo.fits')

    def test_bracket_replace_before(self):
         self.terminal_replace("BRACKET", 1.2, 'foo.fits')
    def test_bracket_replace_mid(self):
         self.terminal_replace("BRACKET", 1.5, 'foo.fits')
    def test_bracket_replace_after(self):
         self.terminal_replace("BRACKET", 5.0, 'foo.fits')

    """
            1.2 : "cref_flatfield_120.fits",
            1.5 : "cref_flatfield_124.fits",
            5.0 : "cref_flatfield_137.fits",
    """
    def test_geometrically_nearest_insert_before(self):
         self.terminal_insert("GEOMETRICALLY_NEAREST", 1.0, 'foo.fits')
    def test_geometrically_nearest_insert_mid(self):
         self.terminal_insert("GEOMETRICALLY_NEAREST", 1.3, 'foo.fits')
    def test_geometrically_nearest_insert_after(self):
         self.terminal_insert("GEOMETRICALLY_NEAREST", 5.1, 'foo.fits')

    def test_geometrically_nearest_replace_before(self):
         self.terminal_replace("GEOMETRICALLY_NEAREST", 1.2, 'foo.fits')
    def test_geometrically_nearest_replace_mid(self):
         self.terminal_replace("GEOMETRICALLY_NEAREST", 1.5, 'foo.fits')
    def test_geometrically_nearest_replace_after(self):
         self.terminal_replace("GEOMETRICALLY_NEAREST", 5.0, 'foo.fits')

# =============================================================================

class RecursiveModify:
    """Tests for checking automatic rmap update logic for adding new references."""

    result_filename = None
    rmap_str = None
    insert_header = lookup_header = {
          "MATCH_PAR1" : "MP1",
          "MATCH_PAR2" : "99.9",
          "DATE-OBS" : "2017-04-20",
          "TIME-OBS" : "00:00:00",
          "SW_VERSION" : "1.2",
          "CLOSETIME" : "2017-05-30 00:01:02",
          "BRACKET_PAR" : "4.4",   # try a number
          "PARAMETER" : "2012-09-09 03:07",
          "GEOM_PAR" : "2.7",  # try a string-formatted number
        }

    def test_0_recursive_modify_rmap(self): # , header, value, classes):
        # Load the test rmap from a string.   The top level selector must exist.
        # This is not a "realistic" test case.   It's a test of the recursive
        # insertion capabilities of all the Selector classes in one go.
        log.verbose("-"*60)
        r = rmap.ReferenceMapping.from_string(self.rmap_str, "./test.rmap", ignore_checksum=True)
        log.verbose("insert_header:", log.PP(self.insert_header))
        result = r.insert(self.insert_header, "foo.fits")
        result.write(self.result_filename)
        diffs = r.difference(result)
        log.verbose("diffs:", diffs)
        diffs = [diff for diff in diffs if "Selector" not in diff[-1]]
        assert len(diffs) == 1, "Fewer/more differences than expected: " + repr(diffs)
        log.verbose("recursive insert result rmap:")
        log.verbose(open(self.result_filename).read())

    def test_1_recursive_use_rmap(self):
        r = rmap.load_mapping(self.result_filename)
        result = r.get_best_ref(self.lookup_header)
        log.verbose("recursive lookup result:", result)
        assert result == self.expected_lookup_result, "Recursively generated rmap produced wrong result."

    def test_9_recursive_tear_down(self):
        os.remove(self.result_filename)

class Test_02_DeepRecursiveModify(TobsTestCase, RecursiveModify):
    result_filename = "./recursive_deep.rmap"
    expected_lookup_result = ("foo.fits", "foo.fits")
    rmap_str = '''
header = {
    'derived_from' : 'Hand written 01-15-2013',
    'filekind' : 'TFILEKIND',
    'instrument' : 'TINSTR',
    'mapping' : 'REFERENCE',
    'name' : 'test.rmap',
    'observatory' : 'TOBS',
    'parkey' : (('MATCH_PAR1','MATCH_PAR2'), ('DATE-OBS','TIME-OBS',), ('SW_VERSION',), ('CLOSETIME',), ('BRACKET_PAR',), ('GEOM_PAR',),),
    'sha1sum' : 'd412b94d1af1a0871fe39d7096e65aea1187c3b7',
    'classes' : ('Match','UseAfter','SelectVersion','ClosestTime','Bracket','GeometricallyNearest')
}

selector = Match({
})
'''

class Test_03_RecursiveUseAfter(TobsTestCase, RecursiveModify):
    result_filename = "./recursive_useafter.rmap"
    expected_lookup_result = "foo.fits"
    rmap_str = '''
header = {
    'derived_from' : 'Hand written 01-15-2013',
    'filekind' : 'TFILEKIND',
    'instrument' : 'TINSTR',
    'mapping' : 'REFERENCE',
    'name' : 'test.rmap',
    'observatory' : 'TOBS',
    'parkey' : (('DATE-OBS','TIME-OBS',), ('MATCH_PAR1','MATCH_PAR2'),),
    'sha1sum' : 'd412b94d1af1a0871fe39d7096e65aea1187c3b7',
    'classes' : ('UseAfter','Match',)
}

selector = UseAfter({
    '2015-04-01 01:02:03' : Match({
    }),
    '2017-04-20 01:02:03' : Match({
    }),
    '2018-04-03 01:02:03' : Match({
    }),
})
'''

class Test_04_RecursiveClosestTime(TobsTestCase, RecursiveModify):
    result_filename = "./recursive_closest_time.rmap"
    expected_lookup_result = "foo.fits"
    rmap_str = '''
header = {
    'derived_from' : 'Hand written 01-15-2013',
    'filekind' : 'TFILEKIND',
    'instrument' : 'TINSTR',
    'mapping' : 'REFERENCE',
    'name' : 'test.rmap',
    'observatory' : 'TOBS',
    'parkey' : (('CLOSETIME',), ('MATCH_PAR1','MATCH_PAR2'),),
    'sha1sum' : 'd412b94d1af1a0871fe39d7096e65aea1187c3b7',
    'classes' : ('ClosestTime','Match',)
}

selector = ClosestTime({
    '2015-04-01 01:02:03' : Match({
    }),
    '2017-04-20 01:02:03' : Match({
    }),
    '2018-04-03 01:02:03' : Match({
    }),
})
'''

class Test_05_RecursiveSelectVersion(TobsTestCase, RecursiveModify):
    result_filename = "./recursive_select_version.rmap"
    expected_lookup_result = "foo.fits"
    rmap_str = '''
header = {
    'derived_from' : 'Hand written 01-15-2013',
    'filekind' : 'TFILEKIND',
    'instrument' : 'TINSTR',
    'mapping' : 'REFERENCE',
    'name' : 'test.rmap',
    'observatory' : 'TOBS',
    'parkey' : (('SW_VERSION',), ('MATCH_PAR1','MATCH_PAR2'),),
    'sha1sum' : 'd412b94d1af1a0871fe39d7096e65aea1187c3b7',
    'classes' : ('SelectVersion','Match',)
}

selector = SelectVersion({
    '<3.1': Match({
    }),
    '<5' : Match({
    }),
    'default' : Match({
    }),
})
'''
    insert_header = lookup_header = {
          "MATCH_PAR1" : "MP1",
          "MATCH_PAR2" : "99.9",
          "SW_VERSION" : "1.2",
        }

class Test_06_RecursiveSelectVersion_MatchingVersion(Test_05_RecursiveSelectVersion):
    insert_header = {
          "MATCH_PAR1" : "MP1",
          "MATCH_PAR2" : "99.9",
          "SW_VERSION" : "<3.1",
        }
    lookup_header = {
          "MATCH_PAR1" : "MP1",
          "MATCH_PAR2" : "99.9",
          "SW_VERSION" : "3.0",
        }

class Test_07_RecursiveSelectVersion_DefaultVersion(Test_05_RecursiveSelectVersion):
    insert_header = lookup_header = {
          "MATCH_PAR1" : "MP1",
          "MATCH_PAR2" : "99.9",
          "SW_VERSION" : "default",
        }

class Test_08_RecursiveGeometricallyNearest(TobsTestCase, RecursiveModify):
    result_filename = "./recursive_geometrically_nearest.rmap"
    expected_lookup_result = "foo.fits"
    rmap_str = '''
header = {
    'derived_from' : 'Hand written 01-15-2013',
    'filekind' : 'TFILEKIND',
    'instrument' : 'TINSTR',
    'mapping' : 'REFERENCE',
    'name' : 'test.rmap',
    'observatory' : 'TOBS',
    'parkey' : (('GEOM_PAR',), ('MATCH_PAR1','MATCH_PAR2'),),
    'sha1sum' : 'd412b94d1af1a0871fe39d7096e65aea1187c3b7',
    'classes' : ('GeometricallyNearest','Match',)
}

selector = GeometricallyNearest({
    0.1: Match({
        ('MPX', '98.7') : 'bar.fits',
    }),
    2.8 : Match({
    }),
    99.0 : Match({
    }),
})
'''
    insert_header = lookup_header = {
          "MATCH_PAR1" : "MP1",
          "MATCH_PAR2" : "99.9",
          "GEOM_PAR" : "1.2",
        }
class Test_09_RecursiveGeometricallyNearestExact(Test_08_RecursiveGeometricallyNearest):
    insert_header = lookup_header = {
          "MATCH_PAR1" : "MP1",
          "MATCH_PAR2" : "99.9",
          "GEOM_PAR" : "0.1",
        }

class Test_10_RecursiveBracket(TobsTestCase, RecursiveModify):
    result_filename = "./recursive_bracket.rmap"
    expected_lookup_result = ("foo.fits", "foo.fits")
    rmap_str = '''
header = {
    'derived_from' : 'Hand written 01-15-2013',
    'filekind' : 'TFILEKIND',
    'instrument' : 'TINSTR',
    'mapping' : 'REFERENCE',
    'name' : 'test.rmap',
    'observatory' : 'TOBS',
    'parkey' : (('BRACKET_PAR',), ('MATCH_PAR1','MATCH_PAR2'),),
    'sha1sum' : 'd412b94d1af1a0871fe39d7096e65aea1187c3b7',
    'classes' : ('Bracket','Match',)
}

selector = Bracket({
    0.1: Match({
        ('MPX', '98.7') : 'bar.fits',
    }),
    2.8 : Match({
        ('MPX', '98.7') : 'bar.fits',
    }),
    99.0 : Match({
    }),
})
'''
    insert_header = lookup_header = {
          "MATCH_PAR1" : "MP1",
          "MATCH_PAR2" : "99.9",
          "BRACKET_PAR" : "0.5",
        }

class Test_11_RecursiveBracketExact(Test_10_RecursiveBracket):
    expected_lookup_result = ("foo.fits", "foo.fits")
    insert_header = lookup_header = {
          "MATCH_PAR1" : "MP1",
          "MATCH_PAR2" : "99.9",
          "BRACKET_PAR" : "0.1",
        }

class Test_12_RecursiveBracketExactMidLookup(Test_10_RecursiveBracket):
    expected_lookup_result = ("foo.fits", "bar.fits")
    insert_header = {
          "MATCH_PAR1" : "MP1",
          "MATCH_PAR2" : "99.9",
          "BRACKET_PAR" : "0.1",
        }
    lookup_header = {
          "MATCH_PAR1" : "MP1",
          "MATCH_PAR2" : "99.9",
          "BRACKET_PAR" : "0.5",
        }
    rmap_str = '''
header = {
    'derived_from' : 'Hand written 01-15-2013',
    'filekind' : 'TFILEKIND',
    'instrument' : 'TINSTR',
    'mapping' : 'REFERENCE',
    'name' : 'test.rmap',
    'observatory' : 'TOBS',
    'parkey' : (('BRACKET_PAR',), ('MATCH_PAR1','MATCH_PAR2'),),
    'sha1sum' : 'd412b94d1af1a0871fe39d7096e65aea1187c3b7',
    'classes' : ('Bracket','Match',)
}

selector = Bracket({
    0.1: Match({
        ('MP1', '99.9') : 'bar.fits',
    }),
    2.8 : Match({
        ('MP1', '99.9') : 'bar.fits',
    }),
    99.0 : Match({
    }),
})
'''

class Test_13_DeleteTest(TobsTestCase):
    result_filename = "./delete.rmap"
    lookup_header = {
          "MATCH_PAR1" : "MP1",
          "MATCH_PAR2" : "99.9",
          "BRACKET_PAR" : "0.5",
        }
    rmap_str = '''
header = {
    'derived_from' : 'Hand written 01-15-2013',
    'filekind' : 'TFILEKIND',
    'instrument' : 'TINSTR',
    'mapping' : 'REFERENCE',
    'name' : 'test.rmap',
    'observatory' : 'TOBS',
    'parkey' : (('BRACKET_PAR',), ('MATCH_PAR1','MATCH_PAR2'),),
    'sha1sum' : 'd412b94d1af1a0871fe39d7096e65aea1187c3b7',
    'classes' : ('Bracket','Match',)
}

selector = Bracket({
    0.1: Match({
        ('MP1', '99.9') : 'bar.fits',
    }),
    2.8 : Match({
        ('MP1', '99.9') : 'bar.fits',
    }),
    99.0 : Match({
    }),
})
'''
    def test_0_recursive_modify_rmap(self): # , header, value, classes):
        # Load the test rmap from a string.   The top level selector must exist.
        # This is not a "realistic" test case.   It's a test of the recursive
        # insertion capabilities of all the Selector classes in one go.
        log.verbose("-"*60)
        r = rmap.ReferenceMapping.from_string(self.rmap_str, "./test.rmap", ignore_checksum=True)
        result = r.delete("bar.fits")
        result.write(self.result_filename)
        log.verbose("result:\n", str(result))
        diffs = r.difference(result)
        log.verbose("diffs:", diffs)
        diffs = [diff for diff in diffs if "Selector" not in diff[-1]]
        assert len(diffs) == 2, "Fewer/more differences than expected: " + repr(diffs)
        for diff in diffs:
            assert "deleted Bracket rule" in diff[-1], "Bad difference " + repr(diff)
        log.verbose("recursive delete result rmap:")
        log.verbose(open(self.result_filename).read())

    def test_1_recursive_use_rmap(self):
        r = rmap.load_mapping(self.result_filename)
        ref = r.get_best_ref(self.lookup_header)
        log.verbose("ref:", ref)
        assert ref.startswith("NOT FOUND list index out of range")

    def test_2_delete_fails(self):
        log.verbose("-"*60)
        r = rmap.ReferenceMapping.from_string(self.rmap_str, "./test.rmap", ignore_checksum=True)
        try:
            result = r.delete("shazaam.fits")
        except crds.CrdsError:
            pass
        else:
            assert False, "Expected delete to fail."


    def test_9_recursive_tear_down(self):
        os.remove(self.result_filename)


if __name__ == '__main__':
    # log.set_verbose()
    unittest.main()



"""This module is used to verify the availability of a list of CRDS files
at the archive web server.
"""
import os.path
import sys

from crds.core import log, config, utils, cmdline
from crds.client import api

DISABLED = []
try:
    import requests
except (ImportError, RuntimeError):
    log.verbose_warning("Failed to import 'requests' module.  check_archive disabled.")
    DISABLED.append("requests")

class CheckArchiveScript(cmdline.Script):
    """Command line script for for checking archive file availability."""

    description =  """Command line script for for checking archive file availability."""

    epilog = """
Checking out the archive with respect to CRDS file availability
and setting up a pipeline cache might have a few stages:

1. Run crds.list to generate a list of files to check, for example for
all contexts, do this:

    % setenv CRDS_SERVER_URL https://jwst-crds.stsci.edu
    % setenv CRDS_PATH $HOME/crds_cache_ops
    % crds list --all --mappings --references >files.b6

This captures the file list with respect to the CRDS OPS server which
during the development period directly distributes files and is a
complete "reference" copy.

2. Run crds.misc.check_archive to query the archive server with HTTP HEAD for each of the files:

    % setenv CRDS_SERVER_URL https://jwst-crds-b6it.stsci.edu
    % setenv CRDS_PATH $HOME/crds_cache_b6it
    crds check_archive --files @files.b6 --dump-good-files --stats > all.b6it.archive

This checks the file list in files.b6 against the CRDS B6 server
database and whatever archive URL that server is configured to use,
nominally the B6 archive server.

crds.misc.check_archive currently checks each file for availability and
correct length with respect to the CRDS catalog.  A status line for
each bad file, or all files, is printed to standard out, in this case
captured to output file all.b6.archive.  e.g. status like:

    jwst_nirspec_superbias_0001.rmap 15a45e1ef114b1d5dc2017fd847830fd3c3eaac6 518 ok
    jwst_nirspec_superbias_0002.rmap 07c94beb372396253dd42c2ff4f2c54ac678d55c 596 ok
    jwst_nirspec_superbias_0003.rmap a031720e6ffeec9d730ec0051a6cea1215e7054f 666 ok

ERROR messages for each problem (missing or length) are printed to stderr

3. Once the archive is looking good,  sync all the CRDS rules to a local cache:

    % crds sync --all --stats --check-sha1sum

which will verify rules download, loadability, and exact contents with respect to
the CRDS server database.

4. Once all of these trial runs are completed,  the appropriate archive pipeline
is probably ready to run the cron_sync script to initialize their CRDS cache from
the archive and appropriate CRDS server.
    """

    def __init__(self, *args, **keys):
        if DISABLED:
            log.fatal_error("Missing or broken dependencies:", DISABLED)
        super(CheckArchiveScript, self).__init__(*args, **keys)
        self.file_info = {}
        self.missing_files = []
        self.bad_length_files = []

    def add_args(self):
        """Add additional custom parameter for CheckArchiveScript."""
        self.add_argument('--files', nargs='*', dest='files', default=[],
                          help='names of files to check for archive availability.')
        self.add_argument('--dump-good-files', action='store_true',
                          help='dump info on good files as well as those with errors.')
        super(CheckArchiveScript, self).add_args()

    @property
    def files(self):
        """Return the filename basenames specified by --files parameters or @-file."""
        return [ os.path.basename(x) for x in self.get_files(self.args.files) ]

    @property
    def mapping_url(self):
        """Return the base archive URL associated with mappings."""
        return self.server_info["mapping_url"][self.observatory]

    @property
    def reference_url(self):
        """Return the base archive URL associated with references."""
        return self.server_info["reference_url"][self.observatory]

    def main(self):
        """Check files for availability from the archive."""
        self.require_server_connection()
        log.info("Mapping URL:", repr(self.mapping_url))
        log.info("Reference URL:", repr(self.reference_url))
        stats = utils.TimingStats()
        self.init_files(self.files)
        for filename in self.files:
            self.verify_archive_file(filename)
            stats.increment("files")
        self.print_files()
        stats.report_stat("files")
        log.standard_status()

    def init_files(self, files):
        self.file_info = api.get_file_info_map(self.observatory, files, fields=["size", "sha1sum"])

    def archive_url(self, filename):
        """Return the URL used to fetch `filename` from the archive."""
        if config.is_mapping(filename):
            return os.path.join(self.mapping_url, filename)
        else:
            return os.path.join(self.reference_url, filename)

    def verify_archive_file(self, filename):
        """Verify the likely presence of `filename` on the archive web server.  Issue an ERROR if absent."""
        url = self.archive_url(filename)
        response = requests.head(url)
        if response.status_code in [200,]:
            log.verbose("File", repr(filename), "is available from", repr(url))
            return self.check_length(filename, response)
        else:
            log.error("File", repr(filename), "failed HTTP HEAD with code =", response.status_code, "from", repr(url))
            self.missing_files.append(filename)
            return False

    def check_length(self, filename, response):
        """Check the content-length reported by HEAD against the CRDS database's file size."""
        archive_size = int(response.headers["content-length"])
        crds_size = int(self.file_info[filename]["size"])
        if archive_size != crds_size:
            log.error("File", repr(filename), "available but length bad.  crds size:", crds_size,
                      "archive size:", archive_size)
            self.bad_length_files.append(filename)
            return False
        else:
            log.verbose("File", repr(filename), "lengths agree:", crds_size)
            return True

    def print_files(self):
        """Print out info on all missing or bad files."""
        for filename in self.missing_files:
            self.dump_file(filename, "bad_missing")
        for filename in self.bad_length_files:
            self.dump_file(filename, "bad_length")
        if self.args.dump_good_files:
            for filename in self.files:
                if filename not in self.missing_files + self.bad_length_files:
                    self.dump_file(filename, "ok")

    def dump_file(self, filename, kind):
        """Dump info about one file annotated with string `kind`."""
        if self.file_info[filename] != "NOT FOUND":
            print(filename, self.file_info[filename]["sha1sum"], self.file_info[filename]["size"], kind)

def file_available(filename):
    """Return True IFF `filename` is believed to be available,  nominally
    based on HTTP HEAD to the archive.
    """
    with log.error_on_exception("Failed verify_archive_file() for",
                                repr(filename)):
        script = CheckArchiveScript()
        script.init_files([filename])
        available = script.verify_archive_file(filename)
        return available

if __name__ == "__main__":
    sys.exit(CheckArchiveScript()())



"""
Special-purpose validators used only with HST synphot files.
"""
import re
import os

import numpy as np

import crds
from crds import data_file
from crds.core import log
from crds.misc.synphot import utils

from .core import Validator


class SynphotGraphValidator(Validator):
    """
    Custom validations for synphot graph table (tmgtab) files.
    """
    def check_header(self, filename, header):
        with data_file.fits_open(filename) as hdul:
            result = self._check_connectivity(hdul)
            result = self._check_direction(hdul) and result
            return result

    def _check_connectivity(self, hdul):
        """
        Confirm that all rows in the graph can be reached by following paths
        starting at innode = 1.
        """

        # This is inefficient, since the subpaths are often shared,
        # but graphs don't get certified regularly, and it seems worthwhile
        # to sacrifice a little speed to keep the code simple.
        def _get_visited_indexes(graph, innode=1, seen_nodes=None):
            if seen_nodes is None:
                seen_nodes = set()

            if innode in seen_nodes:
                # Cycles will result in an error in _check_direction, so
                # we don't need to log an error here.
                return set()

            seen_nodes = seen_nodes.union({innode})

            selected = graph["INNODE"] == innode

            visited = set(np.where(selected)[0])

            for outnode in np.unique(graph[selected]["OUTNODE"]):
                visited.update(_get_visited_indexes(graph, outnode, seen_nodes=seen_nodes))

            return visited

        graph = hdul[-1].data
        visited_indexes = _get_visited_indexes(graph)

        result = True
        if len(visited_indexes) < len(graph):
            missing_indexes = set(range(len(graph))) - visited_indexes
            log.error(
                "TMG contains disconnected rows at indexes:",
                ", ".join(str(i) for i in missing_indexes)
            )
            result = False

        return result

    def _check_direction(self, hdul):
        """
        Confirm that all rows in the graph point to higher INNODE values
        than themselves, which prevents cycles.
        """
        graph = hdul[-1].data
        result = True
        if not (graph["INNODE"] < graph["OUTNODE"]).all():
            log.error(
                "TMG contains rows with INNODE >= OUTNODE."
            )
            result = False

        return result


class SynphotLookupValidator(Validator):
    """
    Custom validations for synphot throughput component lookup
    table (tmctab) and thermal component lookup (tmctab) files.
    """
    def check_header(self, filename, header):
        with data_file.fits_open(filename) as hdul:
            return self._check_filenames(hdul)

    def _check_filenames(self, hdul):
        """
        Confirm that values in the FILENAME column are prefixed
        with a valid path variable, and are suffixed with the correct
        parametrization variable.
        """
        result = True
        for i, row in enumerate(hdul[-1].data):
            expected_path_prefix = utils.get_path_prefix(row["COMPNAME"])
            if row["FILENAME"].split("$")[0] + "$" != expected_path_prefix:
                log.error("Malformed FILENAME value at index", i, "(missing or invalid path prefix)")
                result = False

            param_keyword = utils.get_parametrization_keyword(row["COMPNAME"])
            if param_keyword is None and row["FILENAME"].endswith("]"):
                log.error("Malformed FILENAME value at index", i, "(should not be parametrized)")
                result = False
            elif param_keyword is not None and not row["FILENAME"].lower().endswith("[{}]".format(param_keyword)):
                log.error("Malformed FILENAME value at index", i, "(should be parametrized)")
                result = False

        return result


class SynphotThroughputValidator(Validator):
    """
    Custom validations for synphot component throughput table (thruput) files.
    """
    def check_header(self, filename, header):
        with data_file.fits_open(filename) as hdul:
            result = self._check_throughput_first_and_last(hdul)
            result = self._check_parametrization(hdul) and result
            result = _check_component_filename(
                self.context, utils.THROUGHPUT_REFTYPE, utils.THROUGHPUT_FILENAME_SUFFIX,
                filename, header
            ) and result
            return result

    def _check_throughput_first_and_last(self, hdul):
        """
        Confirm that the first and last values of THROUGHPUT and MJD# columns
        is zero.  Emit a warning only, since there are exceptions to this rule.
        """
        for column_name in hdul[-1].data.names:
            if column_name == "THROUGHPUT" or column_name.startswith("MJD#"):
                if not (hdul[-1].data[column_name][0] == 0.0):
                    log.warning("First value of column '{}' is not 0.0".format(column_name))
                if not (hdul[-1].data[column_name][-1] == 0.0):
                    log.warning("Last value of column '{}' is not 0.0".format(column_name))

        return True

    def _check_parametrization(self, hdul):
        """
        If the component table is parametrized, confirm that it has at least 2
        parametrized columns.
        """
        component = hdul[0].header["COMPNAME"]

        column_prefix = utils.get_parametrization_keyword(component)
        if column_prefix is not None:
            column_count = len([n for n in hdul[-1].data.names if n.lower().startswith(column_prefix)])
            if column_count < 2:
                template = "Table is parametrized by {}, but includes only {} columns with that prefix."
                log.error(template.format(column_prefix, column_count))
                return False

        return True


class SynphotThermalValidator(Validator):
    """
    Custom validations for synphot component thermal table (thermal) files.
    """
    def check_header(self, filename, header):
        return _check_component_filename(
            self.context, utils.THERMAL_REFTYPE, utils.THERMAL_FILENAME_SUFFIX,
            filename, header
        )


def _check_component_filename(context, reftype, suffix, filename, header):
    compname = header.get("COMPNAME", "").lower()
    if not compname:
        log.error("Header missing COMPNAME, unable to certify filename")
        return False

    pattern = re.compile(".*_([0-9]{3})_" + suffix + ".fits")

    match = pattern.match(os.path.basename(filename))
    if not match:
        log.error("Invalid filename: must end in _###_{}.fits".format(suffix))
        return False

    if context:
        new_version = int(match.group(1))

        original_filename = _get_synphot_filename(context, reftype, compname)
        if original_filename:
            match = pattern.match(os.path.basename(original_filename))
            if not match:
                log.warning(
                    "Previous filename '{}' does not end in _###_{}.fits, skipping version check".format(
                        original_filename, suffix
                    )
                )
                return True

            original_version = int(match.group(1))

            if new_version <= original_version:
                log.error(
                    "New filename version ({:03d}) must exceed previous version ({:03d})".format(
                        new_version, original_version
                    )
                )
                return False
        else:
            log.warning(
                "No previous file exists for COMPNAME '{}', skipping version check".format(compname)
            )

    return True


def _get_synphot_filename(context, reftype, compname):
    pmap = crds.get_pickled_mapping(context, ignore_checksum=True)
    rmap = pmap.get_imap(utils.SYNPHOT_INSTRUMENT).get_rmap(reftype)

    for key, filename in rmap.selector.todict()["selections"]:
        if key[0].lower() == compname:
            return filename

    return None



#! /usr/bin/env python
import sys
import glob

from distutils.core import setup

import setuptools

# Define data files included by each observatory package relative
# to each observatory/mission directory.
STD_MISSION_FILES = [
    '*.dat',
    '*.yaml',
    '*.json',
    'tpns/*.tpn',
    'tpns/includes/*.tpn',
    'specs/*.spec',
    'specs/*.rmap',
    'specs/*.json',
]

setup_pars = {
    "packages" : [
        'crds',
        'crds.bestrefs',
        'crds.certify',
        'crds.certify.validators',
        'crds.client',
        'crds.core',
        'crds.io',
        'crds.submit',

        'crds.misc',
        'crds.misc.synphot',
        'crds.refactoring',

        'crds.hst',
        'crds.jwst',
        'crds.tobs',
        'crds.roman',

        'crds.tests',
        ],
    "package_dir" : {
        'crds' : 'crds',
        'crds.bestrefs' : 'crds/bestrefs',
        'crds.certify' : 'crds/certify',
        'crds.certify.validators' : 'crds/certify/validators',
        'crds.client' : 'crds/client',
        'crds.core' : 'crds/core',
        'crds.io' : 'crds/io',
        'crds.submit' : 'crds/submit',

        'crds.misc' : 'crds/misc',
        'crds.misc.synphot' : 'crds/misc/synphot',
        'crds.refactoring' : 'crds/refactoring',

        'crds.hst' : 'crds/hst',
        'crds.jwst' : 'crds/jwst',
        'crds.roman' : 'crds/roman',
        'crds.tobs' : 'crds/tobs',

        'crds.tests' : 'crds/tests',
        },
    "package_data" : {
        'crds.hst': STD_MISSION_FILES,
        'crds.jwst': STD_MISSION_FILES,
        'crds.roman': STD_MISSION_FILES,
        'crds.tobs': STD_MISSION_FILES,
        },
    "scripts" : glob.glob("scripts/*"),
    }

TEST_DEPS = ["lockfile", "mock", "nose", "pytest", "pylint", "flake8", "bandit",]

SUBMISSION_DEPS = ["requests", "lxml", "parsley"]

setup(name="crds",
      provides=["crds"],
      use_scm_version={"write_to": "crds/_version.py"},
      description="Calibration Reference Data System,  HST/JWST/Roman reference file management",
      long_description=open('README.rst').read(),
      author="STScI CRDS s/w developers",
      url="https://hst-crds.stsci.edu",
      license="BSD",
      python_requires=">=3.7",
      setup_requires=["setuptools_scm"],
      install_requires=["astropy", "numpy", "filelock", "asdf"] + SUBMISSION_DEPS,
      extras_require={
          "jwst": ["jwst"],
          "roman" : ["roman_datamodels"],
          "submission": ["requests", "lxml", "parsley"],
          "dev" : ["ipython","jupyterlab","ansible","helm",
                   "nose-cprof", "coverage"],
          "test" : TEST_DEPS,
          "docs" : ["sphinx","sphinx_rtd_theme","docutils","sphinx_automodapi"],
          "aws" : ["boto3","awscli"],
          "synphot": ["stsynphot", "pysynphot"],
      },
      tests_require=TEST_DEPS,
      zip_safe=False,
      classifiers=[
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: BSD License',
          'Operating System :: POSIX :: Linux',
          'Operating System :: MacOS :: MacOS X',
          'Programming Language :: Python :: 3',
          'Topic :: Scientific/Engineering :: Astronomy',
      ],
      project_urls={
          'Documentation' : 'https://hst-crds.stsci.edu/static/users_guide/index.html',
          'Bug Reports': 'https://github.com/spacetelescope/crds/issues/',
          'Source': 'https://github.com/spacetelescope/crds/',
          'Help': 'https://hsthelp.stsci.edu/',
      },
      **setup_pars
)



"""This module is used to profile getrecommendations() and bestreds.BestrefsScript()."""

import pickle

import crds

from crds.tests.test_config import run_and_profile
from crds import data_file

def pickle_unpickle(context, data):
    p = crds.get_cached_mapping(context)
    p.get_required_parkeys()
    header = data_file.get_header(data)
    prefs = p.get_best_references(header)
    q = pickle.dumps(p)
    r = pickle.loads(q)
    rrefs = r.get_best_references(header)
    diffs = p.difference(r, include_header_diffs=True, recurse_added_deleted=True)
    assert prefs == rrefs
    assert str(p) == str(r)
    return {
        "refs": prefs == rrefs,
        "strs" : str(p) == str(r),
        "equal": p == r,
        "diffs": diffs,
        }

if __name__ == "__main__":
    run_and_profile("HST pickle/unpickle",  "pickle_unpickle('hst.pmap', 'data/j8bt06o6q_raw.fits')", globals())



"""Common imports and defs across all versions of ACS hooks.
"""
import sys
from collections import defaultdict

from crds.core import rmap, log, utils, timestamp

# ===========================================================================

#   This section contains matching customizations.

ACS_HALF_CHIP_COLS = 2048         #used in custom bias selection algorithm

SM4 = timestamp.reformat_date("2009-05-14 00:00:00.000000")
# date beyond which an exposure was
# taken in the SM4 configuration
# (day 2009.134 = May 14 2009,
#  after HST was captured by
#  the shuttle during SM4, and
#  pre-SM4 exposures had ceased)
